import logging
from pathlib import Path
from typing import *
from datetime import datetime as dt
import webbrowser
import tkinter as tk


from contextlib import suppress
from tkinter.filedialog import asksaveasfile, askopenfile

import ttkbootstrap as ttk
from ttkbootstrap import colorutils, style as ttk_style
from mdi_pil import ttkbootstrap_mdi as ttk_mdi

from PIL import Image, ImageTk, ImageOps

from inkBoard import core as CORE, packaging
from inkBoard.helpers import QuitInkboard
##Suspect this one won't work here.
##Seems to at least with quit? Check with screenshot too though.

from .. import const
from ..const import THEME_DARK, THEME_LIGHT
from ..settings import EM_SETTINGS

if TYPE_CHECKING:
    from .widgets import DesignerWindow
    from PythonScreenStackManager.elements import Element

    from . import windows, widgets
    from .. import runners

window: "DesignerWindow"

_LOGGER = logging.getLogger(__name__)

def configure_themes(themes : tuple = (THEME_DARK, THEME_LIGHT)):
    "Configures the passed themes to use with inkBoard. Last passed theme will be the one used at startup."

    ibHex = const.INKBOARD_COLORS["inkboard"]
    ibHex = colorutils.color_to_hex(ibHex[:3])
    ibHex_light = const.INKBOARD_COLORS["inkboard-light"]
    ibHex_light = colorutils.color_to_hex(ibHex_light[:3])

    ibHex_grey = const.INKBOARD_COLORS["inkboard-gray"]
    ibHex_grey = colorutils.color_to_hex(ibHex_grey[:3])
    
    for theme in themes:
        style = ttk.Style(theme=theme)
        style.colors.set("dark", ibHex)
        if theme == THEME_LIGHT:
            style.colors.set("primary", ibHex)
            bghsl = colorutils.color_to_hsl(color=style.colors.bg, model="hex")
            newl = min( int(bghsl[-1]*0.85), 100)
                            
            buttonbg = colorutils.update_hsl_value(color=bghsl,lum=newl, inmodel="hsl", outmodel="hex")
            newl = min( int(newl*1.111), 100)
            buttonAct = colorutils.update_hsl_value(color=bghsl,lum=newl, inmodel="hsl", outmodel="hex")
        elif theme == THEME_DARK:
            style.colors.set("primary", ibHex_light)
            bghsl = colorutils.color_to_hsl(color=style.colors.bg, model="hex")
            newl = min( int(bghsl[-1]*1.5), 100)
            
            buttonbg = colorutils.update_hsl_value(color=bghsl,lum=newl, inmodel="hsl", outmodel="hex")
            newl = min( int(newl*0.9), 100)
            buttonAct = colorutils.update_hsl_value(color=bghsl,lum=newl, inmodel="hsl", outmodel="hex")
            
            ##Putting this outside the dark theme config somehow messes up the generation (since the bars etc. are images made with PIL)
            ##So keep this in here for the correct scrollbar style
            bg = style.colors.bg
            style.configure(const.SCROLLBAR_STYLE, 
                            relief="flat",
                            lightcolor=bg, darkcolor=bg,
                            troughcolor=bg, bordercolor=bg,
                            background=bg,
                            arrowsize=11)
            s = style.configure(const.SCROLLBAR_STYLE)
            style.colors.set("dark", ibHex)

        style.colors.set("secondary", buttonbg)
        style.configure('main.TFrame', relief='flat', background=style.colors.bg)

        ##Relief must be flat otherwise there's a white border on the border
        style.configure('right.TFrame', relief='flat', background=ibHex)
        style.configure('buttons.TFrame', relief='flat', background=style.colors.bg)

        ##ttk treeviews seem to not have header borders so will use reliefs
        style.layout('Treeview', 
            [('Treeview.field', {'sticky': 'nswe', 'children': [
                ('custom.Treeview.padding', {'sticky': 'nswe', 'children': [
                    ('custom.Treeview.treearea', {'sticky': 'nswe'})
                    ]})
                ]})
            ])   

        style.configure(f'{const.TREEVIEW_STYLE}.Heading', font=('Arial Bold', 10),                 
            foreground=style.colors.primary,
            background=buttonAct,
            bordercolor=buttonbg,
            darkcolor=buttonAct,
            lightcolor=buttonAct,
            borderwidth=0,
            relief='raised')

        style.map(const.TREEVIEW_STYLE,
            fieldbackground=[('!disabled',style.colors.bg), ('!selected',style.colors.bg)],
            bordercolor=[('!disabled',style.colors.bg),('!selected',style.colors.bg)],
            borderwidth=[('!disabled',0),('!selected',0)])

        style.configure(const.TREEVIEW_STYLE, 
                        borderwidth=0, padding=0, indent=5, background=style.colors.bg
                        )

        style.configure(const.BUTTON_STYLE, borderwidth=10, bordercolor=style.colors.bg,
                        background=buttonbg, relief="solid", shiftrelief=3, 
                        foreground="black")
        
        style.map(const.BUTTON_STYLE, background=[('pressed',buttonAct),('active', buttonAct)],
            relief=[('pressed','sunken'),("!pressed", "solid")],
            highlightthickness = [('pressed',0)]
            )
        
        style.configure("image.TButton", borderwidth=0, bordercolor=style.colors.bg, background=style.colors.bg)
        style.map('image.TButton', 
            background=[('pressed',buttonAct),('active', style.colors.bg)],
            relief=[('pressed','sunken'),("!pressed", "solid"), ]
            )

        style.configure("TLabel",background=style.colors.bg,)
        style.configure(const.LOADBAR_LABEL_STYLE, background=style.colors.secondary,)
        for kw in ttk_style.Keywords.COLORS:
            style.configure(f'{kw}.{const.LOADBAR_LABEL_STYLE}', 
                            background=style.colors.secondary, foreground=getattr(style.colors, kw))
            style.configure(f"{kw}.TLabel", foreground=getattr(style.colors, kw))


        style.configure('noframe.TLabelframe', labeloutside=False, padding=(2,10))
        style.configure('noframe.TLabelframe.Label',  foreground=style.colors.primary)
        style.configure('toggleCenter.TLabelframe', labeloutside=False)
        style.configure('toggleCenter.TLabelframe.Label',  foreground=style.colors.primary)

        style.configure("primary.TLabelframe", padding=(2,0), bordercolor=style.colors.bg)
        style.configure("primary.TLabelframe.Label",font=('Arial Bold', 10), foreground=style.colors.primary)
        
        style.configure("secondary.TLabelframe", padding=(2,0), bordercolor=style.colors.bg)
        style.configure("secondary.TLabelframe.Label",font=('Arial Bold', 10), foreground=style.colors.secondary)

    return
        

def change_theme(*args):
    "Changes the theme based on the value of the darkmode variable"
    newVal = window.globalgetvar(const.DARKMODE_VAR_NAME)
    if newVal:
        _LOGGER.verbose("Turning on dark mode")
        new = THEME_DARK
    else:
        _LOGGER.verbose("Turning off dark mode")
        new = THEME_LIGHT

    window.style.theme_use(new)
    window.trace_variable(const.DARKMODE_VAR_NAME)

_BG_IMG_TK = []
def build_canvas_background(size: tuple[int,int]):
    "(re)builds the background for the designer"

    if _BG_IMG_TK:
        _BG_IMG_TK.clear()

    img_file = Path(__file__).parent.parent / "files" / "background_alpha.png"
    bgImg = Image.open(img_file)
    bgImg = ImageOps.pad(bgImg,size, centering=(1, 1)) #Centering here should align the image to the right, and vertically centered
    bgTk = ImageTk.PhotoImage(bgImg)
    _BG_IMG_TK.append(bgTk)
    return bgTk

def make_screenshot():
    "Makes a screenshot of the current dashboard view."
    if not hasattr(CORE.screen.device,"last_printed_PIL"):
        return
    
    img: Image.Image = CORE.screen.device.last_printed_PIL.copy()
    save_image(img)

def save_image(img: Image.Image, base_name: str = "Inkboard_Screenshot_"):
    """
    Saves an image instance, opens the save_as window depending on the setting

    Parameters
    ----------
    img : Image.Image
        The image to save
    base_name : str, optional
        Base name to use for the filename, by default "Inkboard_Screenshot_"
    """    

    date = dt.now().strftime("%Y_%m_%d_%H%M%S")
    filename = str(base_name) + date
    folder = CORE.config.baseFolder / "screenshots"
    if EM_SETTINGS.getboolean(const.SAVEAS_VAR_NAME,True):
        
        if not Path(folder).exists():
            folder.mkdir()

        _LOGGER.debug("Opening save as dialog")
        files = [('PNG', '*.png'), 
            ('JPEG', '*.jpg'),
            ("BMP", "*.bmp"),
            ('All Files', '*.*')] 
        file = asksaveasfile(filetypes = files, defaultextension = files,
                            initialdir=folder, initialfile=filename)
        if file == None:
            return
        filename = file.name
    else:
        
        if not folder.exists():
            folder.mkdir()
        filename = folder / f"{filename}.png"
    
    img.save(filename)
    _LOGGER.info(f"Screenshot saved as {filename}")

def make_package(*args):
    ##Will extend this later to include dealing saveas screens etc.

    if (not hasattr(CORE,"config") 
        or not hasattr(CORE,"device")
        or not hasattr(CORE,"screen")
        ): 
        _LOGGER.warning("Cannot create package without a complete CORE")
        return

    file = asksaveasfile(
        confirmoverwrite=True,
        filetypes=[("ZIP", "*.zip")],
        defaultextension=[("ZIP", "*.zip")],
        initialdir=CORE.config.baseFolder,
        initialfile=f'inkBoard_package_{CORE.device.emulated_platform}_{CORE.config.filePath.stem}'
    )
    if file == None:
        return

    file = Path(file.name)
    filename = file.stem
    folder = file.parent

    packaging.Packager(CORE, folder).create_package(filename)
    return


def open_device_window(event):
    
    ##This maybe does not make much sense?
    device = getattr(CORE,"device",None)

    windows.DeviceWindow(device)
    return

def open_config_window(event):
    windows.ConfigWindow(CORE)
    return


def set_slider_left_click(event : tk.Event):
    """Sets a sliders position when left clicking by simulating a right click. Bind this function to a sliders left click to do so."""
    widg = event.widget
    widg.event_generate('<Button-3>', x=event.x, y=event.y)
    return

def reload_config(*args):
    #Starts the reload process
    if window._inkBoard_thread.is_alive() and hasattr(CORE,"screen"):
        CORE.screen.reload()
    elif window._current_config_file:
        window._mainLoop.create_task(
            runners.reload_config(window._current_config_file))
    else:
        _LOGGER.warning("No config is currently loaded")


def stop_emulator(*args, exce=None, new_state=None):
    try:
        window.set_inkboard_state(new_state)

        if new_state == None:
            window.update()
    except tk.TclError:
        pass

    if not window._inkBoard_thread.is_alive():
        return

    if exce == None:
        exce = QuitInkboard("Quitting Emulator")

    if hasattr(CORE,"screen") and CORE.screen.printing:
        CORE.screen.quit(exce)

    return


MDI_TREE_ICONS = {}
def build_tree_icon(icon: str) -> ttk_mdi.MDIIcon:
    """
    Builds an icon widget for the element tree, or returns it if it already exists

    Parameters
    ----------
    icon : str
        mdi icon to make an icon off

    Returns
    -------
    ttk_mdi.MDIIcon
        The icon widget
    """    
    if icon in MDI_TREE_ICONS:
        return MDI_TREE_ICONS[icon]
    
    mdi_icon = icon
    icon = ttk_mdi.MDIIcon(icon, const.TREEVIEW_ICON_SIZE)
    MDI_TREE_ICONS[mdi_icon] = icon
    return icon

def get_tree_icon(icon: ttk_mdi.mdiType):
    """Gets the icon widget corresponding to the provided mdi icon. If it does not exist, it is made.

    Parameters
    ----------
    icon : ttk_mdi.mdiType
        mdi icon to get the icon of

    Returns
    -------
    ImageTk.PhotoImage
        The object to be used as a treeview item's image argument
    """    
    return build_tree_icon(icon)

def validate_positive_number(value: str, widget_name: str):
    try:
        value = value.lstrip().replace(",",".")
        value = float(value)
    except ValueError:
        return False
        
    w = window.nametowidget(widget_name)

    if value >= 0:
        w.set(str(value))
        ##This retrieves the actual command attached to the widget
        cmd = lambda : window.tk.call(w['command'])
        cmd()
        return True
    else:
        return False

ELEMENT_ICONS_MDI = {"default": const.DEFAULT_ELEMENT_ICON}
ELEMENT_ICONS_TK = {}

def get_element_tree_icon(element: "Element"):
    
    if "default" not in ELEMENT_ICONS_TK:
        ELEMENT_ICONS_TK["default"] = build_tree_icon(ELEMENT_ICONS_MDI["default"])

    elt_class = str(element.__class__.__name__)
    elt_icon = element._emulator_icon
    if ELEMENT_ICONS_MDI.get(elt_class,"not present") != elt_icon:
        icon = build_tree_icon(elt_icon)
        ELEMENT_ICONS_MDI[elt_class] = elt_icon
        ELEMENT_ICONS_TK[elt_class] = icon
        return icon
    
    return ELEMENT_ICONS_TK.get(elt_class,ELEMENT_ICONS_TK["default"])

##Move highlight function to here.
_INDICATOR_RECTANGLES = []
def highlight_elements(*element_list : "Element"):
    "Pass items to draw a square around. Removes currently drawn squares first."
    _LOGGER.verbose(f"Removing {len(_INDICATOR_RECTANGLES)} rectangles.")
    for rect in _INDICATOR_RECTANGLES:
        ##If this throws errors after clearing: either use find_above and just delete them from there
        ##Just need to know the tag of the screen image itself.
        window.screenCanvas.delete(rect)

    _INDICATOR_RECTANGLES.clear()

    if not EM_SETTINGS.getboolean(const.HIGHLIGHT_VAR_NAME):
        return

    for elt in element_list:
        _LOGGER.debug(f"Selected element in area {elt.area}")
        if elt.area == None or not elt.onScreen:
            continue
        [(x,y),(w,h)] = elt.area
        rect = window.screenCanvas.create_rectangle(x,y,x+w,y+h, outline=const.HIGHLIGHT_COLOR,
                                    dash=const.HIGHLIGHT_DASH, width=const.HIGHLIGHT_WIDTH)
        _INDICATOR_RECTANGLES.append(rect)    

def open_config_folder(*args):
    with suppress(AttributeError):  #Suppressing in case it's called too early e.g.
        folder = CORE.config.baseFolder ##Better to use the folder, using the file seemed to open it in vscode
        webbrowser.open("file:///" + str(folder))
    

    ##Current solution (using webbrowerser) found here: https://stackoverflow.com/a/54641180
    ##Should work regardless of platform?
    ##Otherwise, try e.g. this package: https://pypi.org/project/show-in-file-manager/

    return

def open_new_config(*args):
    
    if hasattr(CORE,"config"):
        init_folder = CORE.config.baseFolder
    elif window._current_config_file:
        init_folder = Path(window._current_config_file).parent.absolute()
    else:
        init_folder = Path.cwd()

    _LOGGER.debug("Opening save as dialog")
    files = [('YAML', '.yaml .yml'),
        ('All Files', '*.*')] 
    file_obj = askopenfile(filetypes = files, defaultextension = files,
                        initialdir=init_folder, )
    
    if not file_obj: 
        return
    
    file = file_obj.name
    
    window._mainLoop.create_task(
        runners.run_inkboard_config(file))
    
    return


"""Kobo device for working with only PSSM (not inkBoard)
"""

#!/usr/bin/env python
import sys
import os
import subprocess
import socket
import asyncio
import logging
import concurrent.futures
from typing import *
from pathlib import Path
from math import ceil
from contextlib import suppress

#Fbink functions etc. can best be checked here: https://github.com/NiLuJe/FBInk/blob/master/fbink.h
#But this is all in C, so some translating may be needed -> yawk has an fbink mock

from PythonScreenStackManager import constants as const, devices as basedevice, tools, exceptions as pssm_exceptions, elements
from PythonScreenStackManager.tools import DummyTask, TouchEvent
from PythonScreenStackManager.pssm_types import *

from PIL import Image, ImageFont, ImageOps

from . import aioKIP, util
from .aioKIP import InputQueue
from .fbink import API as FBInk

_LOGGER = logging.getLogger(__name__)

path_to_pssm_device = os.path.dirname(os.path.abspath(__file__))

touchPath 			= "/dev/input/event1"
batteryCapacityFile = "/sys/devices/platform/pmic_battery.1/power_supply/mc13892_bat/capacity"
batteryStatusFile   = "/sys/devices/platform/pmic_battery.1/power_supply/mc13892_bat/status"

original_truetype = ImageFont.truetype

FEATURES = basedevice.FEATURES

def truetype_wrapper(font = None, size: int = 10, index: int = 0, encoding: str = "", layout_engine = None):
	"Wrapper for truetype fonts to accept Path instance as font value"
	if isinstance(font,Path):
		font = str(font)
	return original_truetype(font,size,index,encoding,layout_engine)

ImageFont.truetype = truetype_wrapper

def PIL_cover_wrapper(image: Image.Image, size: tuple[int, int], method: int = Image.BICUBIC):
	"Wrapper function for PIL cover, since it is not available on the fbink provided PIL"

	##this is literally taken from the source of PIL.
	im_ratio = image.width / image.height
	dest_ratio = size[0] / size[1]

	if im_ratio != dest_ratio:
		if im_ratio > dest_ratio:
			new_height = round(image.height / image.width * size[0])
			if new_height != size[1]:
				size = (size[0], new_height)
		else:
			new_width = round(image.width / image.height * size[1])
			if new_width != size[0]:
				size = (new_width, size[1])
	return image.resize(size, resample=method)

ImageOps.cover = PIL_cover_wrapper

class ResamplingWrapper:
	"Wrapper for the PIL resampling class to at least provide the necessary constants"
	NEAREST = Image.NEAREST
	BOX = Image.BOX
	BILINEAR = Image.BILINEAR
	HAMMING = Image.HAMMING
	BICUBIC = Image.BICUBIC
	LANCZOS = Image.LANCZOS

Image.Resampling = ResamplingWrapper

feature_list = [FEATURES.FEATURE_INTERACTIVE, FEATURES.FEATURE_BACKLIGHT, FEATURES.FEATURE_BATTERY, FEATURES.FEATURE_NETWORK, FEATURES.FEATURE_POWER]
full_device_name = f"{FBInk.platform} {FBInk.device_name}"

class Device(basedevice.PSSMdevice):

	def __init__(self, name: str = full_device_name, rotation: RotationValues = "UR", kill_os: bool = True,
			touch_debounce_time: DurationType = aioKIP.DEFAULT_DEBOUNCE_TIME, hold_touch_time: DurationType = aioKIP.DEFAULT_HOLD_TIME, input_device_path: str = aioKIP.DEFAULT_INPUT_DEVICE):
		"""A base device to run with PSSM. Importing applies some fixes to PIL as well.

		There is support for long touches, however the input library is unable to descern the coordinates of the initial touch.

		Parameters
		----------
		name : str, optional
			The name of the device, by default full_device_name as gotten from fbink
		kill_os : bool, optional
			This kills most of the running kobo processes when the device is initalised, by default True
			This should prevent the device going into sleep mode, for example.
		touch_debounce_time : DurationType, optional
			The default time to allow a full touch to be registered, by default aioKIP.DEFAULT_DEBOUNCE_TIME
		hold_touch_time : DurationType, optional
			The minimum time to hold a touch for it to be passed as being held, by default aioKIP.DEFAULT_HOLD_TIME
		input_device_path : str, optional
			Optionally the path to the touchscreen input, by default aioKIP.DEFAULT_INPUT_DEVICE
		"""	

		features = basedevice.DeviceFeatures(*feature_list)
		
		if kill_os:
			util.kill_os()

		self._battery = Battery()
		self._backlight = Backlight(self)
		self._network = Network()

		super().__init__(features, 
				FBInk.screen_width, FBInk.screen_height, FBInk.view_width, FBInk.view_height,
				"RGBA", "RGBA", "white",
				name)

		self.__KIPargs = {"input_device": input_device_path}
		self.__KIPargs["debounce_time"] = tools.parse_duration_string(touch_debounce_time)
		self.__KIPargs["long_click_time"] = tools.parse_duration_string(hold_touch_time)
		FBInk.rotate_screen(rotation)

	#region
	@property
	def deviceName(self) -> str:
		return full_device_name

	@property
	def screenWidth(self)-> int:
		"Width of the screen"
		return FBInk.screen_width

	@property
	def screenHeight(self)-> int:
		"Height of the screen"
		return FBInk.screen_height

	@property
	def viewWidth(self)-> int:
		"Viewable width of the screen (taking into account possible bezels e.g.)"
		return FBInk.view_width

	@property
	def viewHeight(self)-> int:
		"Viewable height of the screen (taking into account possible bezels e.g.)"
		return FBInk.view_height
	
	@property
	def rotation(self) -> RotationValues:
		i = FBInk.current_rota_canonical
		return get_args(RotationValues)[i]
	
	@property
	def battery(self) -> "Battery":
		return self._battery
	
	@property
	def backlight(self) -> "Backlight":
		return self._backlight
	
	@property
	def network(self) -> "Network":
		return self._network
	#endregion

	def print_pil(self, imgData, x, y, isInverted=False):
		_LOGGER.debug("Printing to device screen")
		if imgData.mode != self.screenMode:
			imgData = imgData.convert(self.screenMode)

		FBInk.fbink_print_pil(imgData,x,y)
		
	async def async_pol_features(self):
		await self.battery.async_update_battery_state()
		await self.network.async_update_network_properties()
		return

	async def event_bindings(self, touch_queue = None):
		self._eventQueue = InputQueue(**self.__KIPargs)
		with suppress(asyncio.CancelledError):
			while True:
				(x,y,action) = await self.eventQueue.get()
				if action == aioKIP.TOUCH_SHORT:
					touch_action = const.TOUCH_TAP
				else:
					touch_action = const.TOUCH_LONG

				await touch_queue.put(TouchEvent(x,y,touch_action))
		return

	def _quit(self, exce=None):
		self._eventQueue.release_input_grab()
		if not isinstance(exce,pssm_exceptions.ReloadWarning):
			self.close_print_handler()

	@staticmethod
	def close_print_handler():
		_LOGGER.info("Closing FBInk")
		FBInk.close()
		
	async def _rotate(self, rotation=None):
		if isinstance(rotation, str):
			rotation = get_args(RotationValues).index(rotation)
		await asyncio.to_thread(FBInk.rotate_screen(rotation))
		await self.Screen._screen_resized()
		await asyncio.to_thread(FBInk.screen_refresh())


	def clear_screen(self):
		"Clears the entire screen"
		FBInk.screen_clear()
	
	def refresh_screen(self, skip_clear: bool = False):
		"Refreshes the entire screen. By default clears it first"
		if not skip_clear:
			self.clear_screen()
		
		FBInk.screen_refresh()

	def set_waveform(self, mode):
		FBInk.set_waveform(mode)

	def reboot(self, *args):
		_LOGGER.info("Rebooting device")
		FBInk.screen_clear()
		FBInk.fbink_print("Rebooting...")
		self.power_off_screen("Rebooting...")
		self.Screen.quit()
		os.system("reboot")

	def power_off(self, *args):
		_LOGGER.info("Powering off device")
		FBInk.screen_clear()
		FBInk.fbink_print("Powering Off")
		self.power_off_screen("Powered Off")
		self.Screen.quit()
		os.system("poweroff")

	def power_off_screen(self, text: str):
		splashBtn = elements.Button(text, text_x_position='left', font_color="white", font="default-bold", font_size=0, fit_text=True)
		splashLayout = [["h*0.7", (None,"w")], ["h*0.2", (None, "?"), (splashBtn,"0.85*w")]]
		img = elements.Layout(splashLayout, background_color="black").generator([(0,0),(self.viewWidth,self.viewHeight)])
		FBInk.fbink_print_pil(img)

# #################### - Hardware etc. - #############################################
class Backlight(basedevice.Backlight):
	'''
	The backlight of the device. Provides callbacks to the state, and functions to turn on, off, or toggle it. Upon initialising this class, the light will be set to 0 to ensure the level is correct
		defaultBrightness (int): default brightness to turn on the backlight too, if not brightness provided (between 1-100)
		defaultTransition (float): default time in seconds for the fade. For smooth fades, 0.5 seems to be the minimum value from my tests.
	'''
	def __init__(self, device: Device, defaultBrightness : int = 50, defaultTransition : float = 0):

		##Ensuring the backlight is off when the dashboard starts, so the brightness and state are correct
		self.__set_backlight_level(0)

		self.__transitionExecutor = concurrent.futures.ThreadPoolExecutor(1,thread_name_prefix="backlightthread")

		super().__init__(device, defaultBrightness, defaultTransition)

	#region
	@property
	def brightness(self) -> int:
		"""The brightness of the backlight (0 - 100)"""
		return self._level

	@property
	def state(self) -> bool:
		"""The state (on/off) of the backlight as a boolean (True/False)"""
		return True if self._level > 0 else False

	@property
	def defaultBrightness(self) -> int:
		"""The default brightness to turn the backlight on to"""
		return self._defaultBrightness

	@defaultBrightness.setter
	def defaultBrightness(self, value : int):
		if value >= 0 and value <= 100:
			self._defaultBrightness = value
		else:
			_LOGGER.error("Default brightness must be between 0 and 100")

	@property
	def defaultTransition(self) -> float:
		"""The default transition time (in seconds)"""
		return self._defaultTransition
	
	@defaultTransition.setter
	def defaultTransition(self, value : float):
		if value >= 0:
			self._defaultTransition = value
		else:
			_LOGGER.error("Default transition time must be 0 or larger")
	#endregion

	async def __set_backlight_level_threadsafe(self, level):
		loop = asyncio.get_running_loop()
		await loop.run_in_executor(
			self.__transitionExecutor,
			self.__set_backlight_level, level)
		##Using a default executor for this seems to improve performance

	def __set_backlight_level(self, level):
		"""
		Args:
			level (int): A frontlight level between 0 (off) and 100 (maximum)
		"""
		if level < 0 or level > 100:
			return
		
		##Do not change this for file reading or whatever. It messes up the file and breaks the frontlight (at least for the python implementation)
		cmd = path_to_pssm_device + "/scripts/frontlight" + f" {level}"
		os.system(cmd)
		self._level = level

	async def __async_transition(self, brightness : int, transition: float):
		"""Async function to provide support for transitions. Does NOT perform sanity checks"""
		if self.brightness == brightness:
			return
		
		async with self._lightLock:
			min_wait = 0.05 ##Chose this value timing using timeit
			wait = transition/(abs(self.brightness-brightness))
			step = -1 if self.brightness > brightness else 1
			if wait < min_wait: 
				steps = ceil(transition/min_wait)
				_LOGGER.debug(f"Fading light from {self.brightness} to {brightness} in {transition} seconds, steps of {steps}")
				for i in range(0,steps):
					step = int(brightness - self.brightness)/(steps-i)
					await self.__set_backlight_level_threadsafe(self.brightness + step)
				await self.__set_backlight_level_threadsafe(brightness)
			else:
				_LOGGER.debug(f"Fading light from {self.brightness} to {brightness} in {transition} seconds. Waiting {wait} seconds between steps")
				while self.brightness != brightness:
					await asyncio.sleep(wait)
					await self.__set_backlight_level_threadsafe(self.brightness + step)
		
		await self.notify_condition()

	async def turn_on_async(self, brightness : int, transition: float):
		"""Async function to provide support for transitions at turn on. Does NOT perform sanity checks"""
		_LOGGER.verbose("Async turning on")
		if self.brightness == brightness:
			##Do nothing if the light is already at the correct level
			return
		
		if transition == None:
			transition = self.defaultTransition

		if brightness == None:
			brightness = self.defaultBrightness
		
		if transition < 0:
			_LOGGER.error("Transition time cannot be negative.")
			return
		
		if brightness < 0 or brightness > 100:
			_LOGGER.error(f"Brightness must be between 0 and 100. {brightness} is an invalid value")
			return
		
		if transition == 0:
			_LOGGER.debug(f"Setting backlight to {brightness} without transition")
			await self.__set_backlight_level_threadsafe(brightness)
			await self.notify_condition()
		else:
			self._transitionTask = asyncio.create_task(self.__async_transition(brightness, transition))
			await self._transitionTask

	def turn_on(self, brightness : int = None, transition : float = None):
		"""Turn on the backlight to the set level"""
		asyncio.create_task(self.turn_on_async(brightness, transition))

	async def turn_off_async(self, transition: float = None):
		"""Async function to provide support for transitions at turn off. Does NOT perform sanity checks"""
		_LOGGER.verbose("Async turning off")
		if not self.state:
			##Do nothing if the light is already off
			return

		if transition == None:
			transition = self.defaultTransition

		if transition < 0:
			_LOGGER.error("Transition time cannot be negative.")
			return

		if transition == 0:
			_LOGGER.debug(f"Turning backlight off without transition")
			self.__set_backlight_level(0)
			await self.notify_condition()
		else:
			self._transitionTask = asyncio.create_task(self.__async_transition(0, transition))
			await self._transitionTask

	def turn_off(self, transition : float = None):
		"""Turns off the backlight to the set level"""
		if not self.state:
			##Backlight is already off, no need to do anything
			return
		
		asyncio.create_task(self.turn_off_async(transition))

	def toggle(self, brightness : int = None, transition : float = None):
		"""Toggles the backlight, if it is off turns on to defined brightness"""
		asyncio.create_task(self.toggle_async(brightness, transition))
	
	async def toggle_async(self, brightness = None, transition = None):
		if transition == None:
			transition = self.defaultTransition

		if transition < 0:
			_LOGGER.error("Transition time cannot be negative.")
			return

		if not self._transitionTask.done():
			self._transitionTask.cancel()

		if self.state:
			await self.turn_off_async(transition)
			return
		else:
			if brightness == None:
				brightness = self.defaultBrightness
			if brightness < 0 or brightness > 100:
				_LOGGER.error(f"Brightness must be between 0 and 100. {brightness} is an invalid value")
				return
			else:
				await self.turn_on_async(brightness, transition)
		
			
class Battery(basedevice.Battery):
	'''
	The battery of the device. Provides callbacks to get the battery state and charge level, as well as update it.
	'''
	def __init__(self):

		##Ensuring the backlight is off when the dashboard starts, so the brightness and state are correct
		charge = self.readBatteryPercentage()
		state = self.readBatteryState()

		self._update_properties((charge,state))

	@property
	def percentage(self):
		"""The battery charge percentage"""
		return self._batteryPercentage
	
	@property
	def state(self):
		"""The state of the battery"""
		return self._batteryState

	async def async_update_battery_state(self):
		charge = await asyncio.to_thread(self.readBatteryPercentage())
		state = await asyncio.to_thread(self.readBatteryState())
		self._update_properties((charge, state.lower()))

	def readBatteryPercentage(self) -> str:
		with open(batteryCapacityFile) as state:
			state.seek(0)
			res = ""
			for line in state:
				res += str(line)		
		return int(res)

	def readBatteryState(self) -> str:
		res=""
		with open(batteryStatusFile) as percentage:
			percentage.seek(0)
			isFirst = True
			for line in percentage:
				if isFirst:
					res += str(line).rstrip()
					isFirst=False
		
		res = res.lower()
		if res == "not charging":
			res = "discharging"
		return res

class Network(basedevice.Network):
	'''
	Handles Network stuff. Gets IP Adress, network SSID etc, and can turn on and off wifi.
	Properties: IP, wifiOn, connected, SSID
	'''
	def __init__(self):
		self._isWifiOn = True
		self._update_network_properties()

	@property
	def IP(self) -> str:
		"""Returns the IP adress"""
		return self._IP

	@property
	def wifiOn(self) -> bool:
		"""Returns whether wifi is on"""
		return self._isWifiOn
	
	@property
	def connected(self) -> bool:
		"""Returns whether the device is connected to a wifi network"""
		return self._connected
	
	@property
	def SSID(self) -> str:
		"""Returns the SSID of the connected network"""
		return self._SSID
	
	@property
	def macAddr(self) -> str:
		"""Returns the mac adress of the device"""
		return self._macAddr
	
	@property
	def signal(self) -> int:
		"Wifi signal percentage, from 0-100, or None if unavailable."
		return None

	async def async_update_network_properties(self):
		await asyncio.to_thread(self._update_network_properties)

	def update_network_properties(self):
		asyncio.create_task(self.async_update_network_properties())

	def _update_network_properties(self):
		self._macAddr = util.get_mac()
		self._connected = util.is_wifi_connected()
		if self.connected:
			self._IP = util.get_ip()
			self._SSID = util.get_SSID()
		else:
			self._IP = None
			self._SSID = None





"""
Holds data for default icon packs for some elements.
"""
import importlib.util
from pathlib import Path
from typing import TYPE_CHECKING

from inkBoard.constants import DESIGNER_INSTALLED
from PythonScreenStackManager.constants import MDI_WEATHER_DATA_ICONS, PATH_TO_PSSM
from mdi_pil import MDI_WEATHER_ICONS as MDI_WEATHER_CONDITION_ICONS

if TYPE_CHECKING:
    from .. import meteocons

METEOCONS_INSTALLED: bool = False
s = importlib.util.find_spec("inkBoard.integrations.meteocons")
if DESIGNER_INSTALLED:
    try:
        s = importlib.util.find_spec("inkBoarddesigner.integrations.meteocons")
    except:
        pass

if s:
    METEOCONS_INSTALLED = True
    METEOCONS = importlib.import_module(s.name)

##Should be able to parse from both inkBoard or designer integration.

HVAC_MODES_ICONS : dict = {
    "off" : "mdi:power",
    "heat": "mdi:fire",
    "cool": "mdi:snowflake" ,
    "heat_cool": "mdi:sun-snowflake-variant" ,
    "dry": "mdi:heat-wave" ,
    "fan_only": "mdi:fan" ,
    "auto": "mdi:thermostat-auto"
}
"Default icons for the possible HVAC modes"


import pkgutil
from pathlib import Path

import ttkbootstrap as ttk

from inkBoard.constants import INKBOARD_COLORS, ARGUMENT_CONFIG
from PythonScreenStackManager.devices.const import CANVASNAME

PSSM_FOLDER = Path(pkgutil.get_loader("PythonScreenStackManager").get_filename()).parent
INKBOARD_FOLDER = Path(pkgutil.get_loader("inkBoard").get_filename()).parent

IMPORTER_THREADPOOL = "designer-importer-threadpool"
BASE_WINDOW_TASK_NAME = "designer-window"

INTEGRATION_DIRS = {
    f"{__package__}.integrations": Path(__file__).parent / "integrations",
    "inkBoard.integrations": INKBOARD_FOLDER / "integrations",
}

##The last one is for the custom integrations. Will be added later on.
##Considering how taking precedent works: probably use the custom folder first.

DEFAULT = "default-designer-value"
"constant to mark default values"

RELOAD_MODULES = [
    "emulator"
]

for i, mod in enumerate(RELOAD_MODULES):
    RELOAD_MODULES[i] = f"{__package__}.{mod}"

HIGHLIGHT_VAR_NAME = "hightlight" ##These have to correspond to the value in the settings
SAVEAS_VAR_NAME = "saveas"
DARKMODE_VAR_NAME = "darkmode"

LIST_VAR_NAME = "treeview-list-variable" ##Maybe this one not but will have to see obviously
ELEMENT_TREE_OPTION = "Elements"
NO_TREE_OPTION = "None"

CANVAS_NAME = CANVASNAME
SCREEN_TAG = "pssm-screen"
UI_FRAME_NAME = "designer-ui-frame"
TREE_FRAME_NAME = "treeview-frame"
HOVER_TAG = "treeview-item-hover"

SCREENSHOT_BUTTON_NAME = "config-button-capture"
SCREENSHOT_BUTTON_TEXT = "Capture"
SCREENSHOT_BUTTON_ICON = "mdi:camera-iris"

PACK_BUTTON_NAME = "config-button-pack"
PACK_BUTTON_TEXT = "Pack"
PACK_BUTTON_ICON = "mdi:folder-download"

RELOAD_BUTTON_NAME = "config-button-reload"
RELOAD_BUTTON_TEXT = "Reload"
RELOAD_BUTTON_ICON = "mdi:sync-circle"

STOP_BUTTON_NAME = "config-button-stop"
STOP_BUTTON_TEXT = "Stop"
STOP_BUTTON_ICON = "mdi:close-circle"

ERROR_ACTIVE_BUTTONS = [STOP_BUTTON_NAME, RELOAD_BUTTON_NAME]
"Buttons that should stay active if loading inkBoard errored"

THEME_DARK = "darkly"
THEME_LIGHT = "yeti"
TREEVIEW_STYLE = "custom.Treeview"
SCROLLBAR_STYLE = "dark.Round.Vertical.TScrollbar"
FEATURE_FRAME_STYLE = ttk.PRIMARY
LOADBAR_LABEL_STYLE = "loadbar.TLabel"

BUTTON_STYLE = "custom.TButton"
TOOLTIP_STYLE = (ttk.DARK, ttk.INVERSE)

INTERACT_CURSOR = "hand2"

ICON_COLOR = INKBOARD_COLORS["inkboard"]
HA_BLUE: tuple = (3, 169, 244, 255)
#Color used for home assistant branding

REFRESH_RATE = 20 #Rate to refresh the window per second (Idk if this updates the canvas too, may indeed update all widgets)

DEFAULT_LABEL_FONT = ('Arial Bold', 10)

INTERFACE_PADDING = 0.04
INTERFACE_BORDER = 5
INTERFACE_WIDTH = 250
BUTTON_HEIGHT = 30
BUTTON_FONT_SIZE = 16
BUTTON_WIDTH = int(INTERFACE_WIDTH*0.36)
BUTTON_ICON_COLOR = ICON_COLOR

SETTINGS_HEIGHT = int(INTERFACE_WIDTH*0.3)
SETTINGS_WIDTH = int(INTERFACE_WIDTH/4)
SETTINGS_PADDING = 5
LIST_WIDTH = 20 #This value is gotten via trial and error, for an INTERFACE_WIDTH of 200; tkinter width's are just kind of an illusion generally
DEFAULT_LIST_OPTIONS = ["None", "Elements"]

HA_FONT_NAME = "Quicksand-bold.ttf"
HA_FONT_FILE = PSSM_FOLDER / "fonts" / HA_FONT_NAME


SCREENSHOT_TIP = "Make a screenshot of the currently shown dashboard"
PACK_TIP = "Make an update/install package of the currently running config (Not implemented)"
RELOAD_TIP = "Reload the configuration and inkBoard instance."
STOP_TIP = "Stop the emulator and close the current config."
SHOW_IMAGE_TIP = "Show the current image of the element in your default image viewer."
SAVE_IMAGE_TIP = "Save the current element image as a file. Follows the value of the 'Save As' setting on whether or not to show a file dialog."
CONFIG_LABEL_TIP = "Open the folder of the current config in your file explorer."
CONFIG_OPEN_TIP = "Choose a config file to run."

DARKMODE_TIP = "Toggle Dark Mode"
HIGHLIGHT_TIP = "Highlight elements in the dashboard when selecting them in the entity or element list"
SAVEAS_TIP = "Show a file explorer window when making a screenshot"
TREE_REFRESH_TIP = "Refresh the current treeview"

CONFIG_OPTIONS_TIP = "Info on the currently opened config (if any), or open a new config."
DEVICE_TIP = "Info on the currently running device (if any), as well as some settings to alter emulation behaviour."

DEFAULT_ELEMENT_ICON = "mdi:shape"
TREEVIEW_ICON_SIZE = (15,15)

HIGHLIGHT_DASH = (5,2)
HIGHLIGHT_WIDTH = 5
HIGHLIGHT_COLOR = "red"


"""Library to interface with fbink in a more convenient manner.
See: https://github.com/NiLuJe/FBInk/blob/master/fbink.h#L621 for the full library
See: https://github.com/NiLuJe/py-fbink/blob/master/fbink_build.py for the available python bindings.

Usage: `from fbink import API as FBInk`
"""

from typing import * 
import functools, os, subprocess, time, logging

if TYPE_CHECKING:
    from .fbink_mock import ffi, lib as FBInk
    from PIL import Image
else:
    from _fbink import ffi, lib as FBInk

VERBOSE = int(logging.DEBUG/2)
_LOGGER = logging.getLogger(__name__)

ffi_typedef = TypeVar("typedef")
ffi_struct = TypeVar("struct")
ffi_union = TypeVar("union")

T = TypeVar('T')
class classproperty(Generic[T]):
    "Used to avoid the deprecation warning (and the extra writing) needed to set class properties"
    
    def __init__(self, method: Callable[..., T]):
        self.method = method
        functools.update_wrapper(self, wrapped=method) # type: ignore

    def __get__(self, obj, cls=None) -> T:
        if cls is None:
            cls = type(obj)
        return self.method(cls)

rotation_map = {"UR": 0, "CW": 1, "UD": 2, "CCW": 3}

class API:
    """Wrapper around the FBInk lib class with doc strings and type hints were possible.

    An API to interface and print stuff on E-ink devices with NiLuJe's Kobo stuff installed, which comes with the FBInk library and python bindings.
    I am by no means an expert, so the implemented methods are only for stuff I understand/needed.
    Upon being garbage collected, the class itself should take care of closing the framebuffer.

    The API has class properties for certain values from the state struct, like the device name and the screen size e.g.
    You can call `API._unused_state_attributes` to get a list of all the unused attributes available in the FBInkState struct, and use `get_state_attribute` to get its value.
    """    

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls,"__instance"):
            cls.__instance = object.__new__(cls)
            return cls.__instance 
        else:
            return cls.__instance
    
    @classmethod
    def __init__(cls):
        cls._fbink_cfg = ffi.new("FBInkConfig *")
        cls._fbink_dumpcfg = ffi.new("FBInkDump *")

        cls._is_quiet = 1
        cls._fbink_cfg.is_quiet = cls._is_quiet
        cls._fbfd = FBInk.fbink_open()
        FBInk.fbink_init(cls._fbfd, cls._fbink_cfg)
        cls._state = ffi.new("FBInkState *")
        cls.get_state()
        v = cls.fbink_version()
        if "for" in v:
            v, plt = v.split(" for ")
        else:
            plt = None
        
        cls._version = v
        cls._platform = plt

    @classmethod
    def __del__(cls):
        FBInk.fbink_close(cls._fbfd)
        print("fbink closed")

    @classmethod
    def close(cls):
        FBInk.fbink_close(cls._fbfd)

    @classmethod
    def get_state(cls):
        cls.fbink_reinit()
        state = cls._state
        FBInk.fbink_get_state(cls._fbink_cfg, state)

        ##For the definition of all attributes of state, see:
        ##https://github.com/NiLuJe/FBInk/blob/master/fbink.h#L621
        for attr in state.__dir__():
            val = getattr(state,attr)
            if isinstance(val, ffi.CData):
                val = ffi.string(val)
                if isinstance(val,bytes):
                    val = val.decode("ascii")
            attr_name = f"_{attr}"
            setattr(cls,attr_name, val)
        cls._state = state
    
    @classmethod
    def get_state_attribute(cls, attribute: str) -> Any:
        "Returns the requested state attributes value."
        if hasattr(cls, f"_{attribute}"):
            return getattr(cls, f"_{attribute}")
        return getattr(cls._state, attribute)
    
    @classmethod
    def _unused_state_attributes(cls):
        """Prints out any properties of the state struct not available as class property
        May be useful when the python bindings are updated (seems to be happening as some c functions like `fbink_target` are unavailable.)
        Though it may simply be due to using an older device, the python bindings have not been updated as recently as the main library, but that may come when a new release of kobo stuff is uploaded.
        """
        
        missing = []
        for attr in cls.__state.__dir__():
            if not hasattr(cls,attr):
                missing.append(attr)
        
        if missing:
            print("FBInk API does not have the following state properties defined:")
            for attr in missing: print(attr)

    @classmethod
    def _all_waveforms(cls) -> list[str]:
        wfms = []
        for attr in FBInk.__dict__:
            if "WFM" in attr:
                wfms.append(attr)
        return wfms
    
    @classmethod
    def _all_methods(cls) -> list[Callable]:
        ms = []
        for attr, val in FBInk.__dict__.items():
            if callable(val):
                ms.append(attr)
        return ms

    @classmethod
    def _ffi_types(cls) -> tuple[list[ffi_typedef],list[ffi_struct],list[ffi_union]]:
        "Returns the available ffi types available"
        return ffi.list_types()

    #region
    @classproperty
    def version(cls) -> str:
        "The fbink version"
        return cls._version
    
    @classproperty
    def platform(cls) -> str:
        "The fbink platform, e.g. Kobo or Kindle. None if unknown"
        return cls._platform

    @classproperty
    def device_name(cls) -> str:
        "The name/model of the device"
        return cls._device_name

    @classproperty
    def device_codename(cls) -> str:
        "Codename of the device"
        return cls._device_codename
    
    @classproperty
    def device_platform(cls) -> str:
        "Platform of the device (board type, not brand name)"
        return cls._device_platform

    @classproperty
    def device_id(cls) -> int:
        "The id of the device, used by fbink to determine the model"
        return cls._device_id

    @classproperty
    def screen_width(cls) -> int:
        "Total width of the screen in pixels"
        return cls._screen_width

    @classproperty
    def screen_height(cls) -> int:
        "Total height of the screen in pixels"
        return cls._screen_height
    
    @classproperty
    def view_width(cls) -> int:
        "Viewable width of the screen in pixels (Taking bezels into account)"
        return cls._view_width
    
    @classproperty
    def view_height(cls) -> int:
        "Viewable height of the screen in pixels (Taking bezels into account)"
        return cls._view_height
    
    @classproperty
    def max_rows(cls) -> int:
        return cls._max_rows
    
    @classproperty
    def max_cols(cls) -> int:
        return cls._max_cols

    @classproperty
    def screen_dpi(cls) -> int:
        "DPI (resolution) of the screen"
        return cls._screen_dpi
    
    @classproperty
    def current_rota(cls) -> int:
        "Current rotation as reported by the device"
        return cls._current_rota

    @classproperty
    def current_rota_canonical(cls):
        "Current rotation converted to canonical rotation (0 for upright portrait)"
        return FBInk.fbink_rota_native_to_canonical(cls.current_rota)

    @classproperty
    def can_rotate(cls) -> bool:
        "True if the device can rotate (i.e. has a gyro)"
        return cls._can_rotate

    @classproperty
    def can_hw_invert(cls) -> bool:
        "True if the device supports hardware screen inversion"
        return cls._can_hw_invert
    #endregion

    #region
    @classmethod
    def fbink_version(cls) -> str:
        "Prints the fbink version"
        return ffi.string(FBInk.fbink_version()).decode("ascii")

    @classmethod
    def fbink_reinit(cls):
        FBInk.fbink_reinit(cls._fbfd, cls._fbink_cfg)

    @classmethod
    def fbink_print(cls, string: str):
        "Prints a string on the top of the screen"
        FBInk.fbink_print(cls._fbfd, string.encode("utf-8"), cls._fbink_cfg)

    @classmethod
    def fbink_refresh(cls, x: int, y: int, w: int, h: int):
        "Refreshes the screen in the provided region"
        FBInk.fbink_refresh(cls._fbfd, y,x,w,h, cls._fbink_cfg)

    @classmethod
    def fbink_print_image(cls, image_file: str, x_off: int, y_off: int):
        "Prints the image file onto the screen"
        FBInk.fbink_print_image(cls._fbfd, bytes(image_file, "utf-8"), x_off, y_off, cls._fbink_cfg)

    @classmethod
    def fbink_print_raw_data(cls, data: str, w: int, h: int, length, x_off: int, y_off: int):
        "Prints the raw data as pixels onto the screen"
        FBInk.fbink_print_raw_data(cls._fbfd, data, w, h, length, x_off, y_off, cls._fbink_cfg)

    @classmethod
    def fbink_print_pil(cls, image: "Image.Image", x: int = 0, y: int = 0):
        """Convenience method to print a PIL image instance to the screen.
        
        Wrapper around `API.fbink_print_raw_data`, as the function decodes the provided image into raw data.

        Parameters
        ----------
        image : Image.Image
            The image object to print
        x : int, optional
            x coordinates of the topleft corner in pixels, by default 0
        y : int, optional
            y coordinates of the topleft corner in pixels, by default 0
        """        
        ""
        if image.mode == "P":
            _LOGGER.debug("Image is paletted, translating to actual values")
            image = image.convert()

        # If image is not grayscale, RGB or RGBA (f.g., might be a CMYK JPEG) convert that to RGBA.
        if image.mode not in ["L", "LA", "RGB", "RGBA"]:
            _LOGGER.debug("Image data is packed in an unsupported mode, converting to RGBA")
            image = image.convert("RGBA")

        # And finally, get that image data as raw packed pixels.
        raw_data = image.tobytes("raw")
        raw_len = len(raw_data)
        _LOGGER.log(VERBOSE,"Raw data buffer length: {}".format(raw_len))

        cls.fbink_print_raw_data(raw_data, image.width, image.height, raw_len, x, y)

    @classmethod
    def fbink_cls(cls, rect):
        "Clears the screen region of the given rectangle pointer"
        FBInk.fbink_cls(cls._fbfd, cls._fbink_cfg, rect)

    @classmethod
    def fbink_grid_clear(cls, cols: int, rows: int):
        "Clears the screen region up to the given column and row"
        FBInk.fbink_grid_clear(cls._fbfd, cols, rows, cls._fbink_cfg)

    @classmethod
    def fbink_grid_refresh(cls, cols, rows):
        "Refreshes the screen up to the given column and row"
        FBInk.fbink_grid_refresh(cls._fbfd, cols, rows, cls._fbink_cfg)
    
    @classmethod
    def fbink_invert_screen(cls):
        "Inverts the screen using software inversion"
        FBInk.fbink_invert_screen(cls._fbfd, cls._fbink_cfg)

    #endregion

    @classmethod
    def screen_clear(cls):
        "Clears the entire screen"
        cls.fbink_grid_clear(cls.max_cols, cls.max_rows)
    
    @classmethod
    def screen_refresh(cls):
        "Refreshes the entire screen"
        cls.fbink_grid_refresh(cls.max_cols, cls.max_rows)

    @classmethod
    def set_waveform(cls, mode: str = "AUTO"):
        if mode == "AUTO":
            mode = "WFM_AUTO"
        try:
            if mode == None:
                raise AttributeError
            mode = mode.upper().replace(" ", "_")
            val = getattr(FBInk,mode)
            cls._fbink_cfg.wfm_mode = val
        except AttributeError:
            print(f"Invalid wafeform value {mode}")
            cls._fbink_cfg.wfm_mode = FBInk.WFM_AUTO

    @classmethod
    def rotate_screen(cls, rota: Union[int,str] = None):
        """Rotates the screen. 
        
        If rota is not passed, will rotate 90 degrees clockwise.
        Passes the command 'fbdepth -R' to the os, since rotation via the Python bindings seemed impossible.
        """

        if isinstance(rota, str):
            if not rota in rotation_map:
                _LOGGER.warning(f"{rota} is not a valid rotation value. Use {rotation_map.keys()}")
                return
            rota = rotation_map[rota]

        if rota == cls.current_rota_canonical:
            return
        
        if rota == None:
            rota = cls.current_rota_canonical + 1 if cls.current_rota_canonical < 3 else 0
        
        if rota > 3:
            _LOGGER.warning("Rotation cannot be larger than 3")
            return
        
        _LOGGER.debug(f"Rotating device from {cls.current_rota_canonical} to {rota}")

        cmd = ["fbdepth", "-R", str(rota)]
        if cls._is_quiet:
            cmd.append("-q")
        subprocess.call(cmd)
        cls.get_state()
        cls.screen_clear()
        cls.screen_refresh()

API()


"""
This device can be used in desktop environments.
It is an extension of the base device that comes with PythonScreenStackManager, since it (will) provide functionality for somewhat easy extension of the functionality using integrations.
"""

import __main__
import asyncio
import tkinter as tk
from typing import TYPE_CHECKING, Union
from pathlib import Path
import platform

from PIL import Image, ImageTk

from inkBoard.platforms.basedevice import BaseDevice, InkboardDeviceFeatures, FEATURES,\
                                    BaseBacklight, BaseBattery, BaseNetwork
from inkBoard.constants import INKBOARD_FOLDER

from PythonScreenStackManager.devices import windowed
from PythonScreenStackManager.tools import DummyTask

try:
    #plyer module is optional, provides additional device features
    import plyer #@IgnoreExceptions
except ModuleNotFoundError:
    plyer = False

if TYPE_CHECKING:
    from PythonScreenStackManager.pssm_types import *

class Device(BaseDevice, windowed.Device):
    """inkBoard device for desktop environments

    This device can be used on any OS platform where python's tkinter library can be used for application windows.
    Some optional features exist, but it can be used without any additional libraries.

    Parameters
    ----------
    name : str, optional
        The name of the device, also used to name the window, by default None
    frame_rate : int, optional
        The amount of times the window is updated per second, by default 20.
        Since this is a dashboarding interface, the value does not need to be very high to still have a decent running experience.
        The main performance bottleneck probably lies in the generating of elements and the fact that it is run in Python.
    width : int, optional
        The initial width of the window, by default 1280
    height : int, optional
        The initial height of the window, by default 720
    fullscreen : bool, optional
        Whether to start the window in fullscreen, by default False
        screenWidth and screenHeight are set to the correct value if this True.
        The device provides a shorthand function for toggling fullscreen, and binds the F11 key to toggle it as well.
    resizeable : bool, optional
        Allows the window to be resized by dragging the edges, by default False
    cursor : str, optional
        The type of cursor to use when the mouse is on the dashboard. By default "target"
        see https://anzeljg.github.io/rin2/book2/2405/docs/tkinter/cursors.html for available types.
    color_mode : str, optional
        The image mode to print the final screen image in, by default "RGB". Use "L" or "LA" if you'd want black and white.
    default_color : ColorType, optional
        The default color (i.e. assumed background color) of the device, by default "white"
    interactive : bool, optional
        Whether the dashboard can be interacted with, by default True
    network : bool, optional
        Whether to use network_features, by default True
        This indicates the device can access the internet and periodically pols the network properties. Setting it to False does not actually block off internet access to the programme.
    backlight : bool, optional
        Whether to simulate screen brightness, i.e. print a transparent black rectangle over the dashboard, by default False. May affect performance when used.
    backlight_alpha : int, optional
        The maximum allowed alpha value of the backlight rectangle, so the transparancy when the backlight is considered off. by default 175
        0 is the minimum value, which is the same as not using the backlight feature. 255 is the maximum value, which means the rectangle is not transparant at all when it is off.
    window_icon : str, Path
        The icon for the window. This must be a .ico file
        Also, on windows, due to how applications are managed, inkBoard will not use this icon in the taskbar.
        Defaults to the 'inkBoard', which means it uses the inkBoard logo icon.
    """  

    def __init__(self, name : str = None,
                frame_rate : int = 20, width = 1280, height = 720, fullscreen : bool = False, resizeable : bool = False, cursor : str = "target",
                color_mode : str = "RGB",  default_color: "ColorType"  = "white",
                interactive : bool = True, network : bool = True, backlight : bool = False, backlight_alpha : int = 175,
                window_icon: Union[str,Path] = "inkboard",
                ):

        self._model = None
        
        feature_dict = {FEATURES.FEATURE_INTERACTIVE: interactive, FEATURES.FEATURE_PRESS_RELEASE: interactive,
                        FEATURES.FEATURE_BACKLIGHT: backlight, FEATURES.FEATURE_NETWORK: network,}
        if plyer:
            bat_state = plyer.battery.get_state()
            if not all(v is None for v in bat_state.values()):
                feature_dict[FEATURES.FEATURE_BATTERY] = True

        features = InkboardDeviceFeatures(**feature_dict)

        if backlight:
            self._backlight = windowed.Backlight(self, backlight_alpha)

        ##Gotta see what needs to be overwritten after this
        windowed.Device.__init__(self, name, frame_rate,  
                        width, height, fullscreen, resizeable,
                        cursor, color_mode, "RGBA", default_color, features=features)
        
        if window_icon in {None, "inkboard"}:
            window_icon = Path(INKBOARD_FOLDER) / "files" / "icons" / "inkboard.ico"
        self.window.wm_iconbitmap(window_icon)

        if name == None:
            self.window.title(f"inkBoard")
        else:
            self.window.title(self.name)

class Battery(BaseBattery):
        "Device battery. Not used if unsupported by plyer."

        def __init__(self, device):
            
            charge, state = self.update_battery_state()

            super().__init__(device, charge, state)

        async def async_update_battery_state(self)-> tuple[int,str]:
            """
            Update the battery state. Returns the result.

            Returns
            -------
            tuple[int,str]
                Tuple with [charge percentage, state]
            """
            ##idk if this is blocking, if so, needs to go to a thread
            self.update_battery_state()
        
        def update_battery_state(self):
            state = plyer.battery.get_state()
            charge_perc = state.get("percentage", 0)

            if state["isCharging"]:
                charge_state = "charging"
            else:
                charge_state = "discharging" if charge_perc < 100 else "full"

            t = (charge_perc, charge_state)
            self._update_properties(t)
            return t
    #endregion




"""
Constants for use with the HA PSSM library
"""
from typing import Literal, Optional, Union
import yaml

from inkBoard import core as CORE

from PythonScreenStackManager.constants import INKBOARD, PSSM_COLORS, SHORTHAND_FONTS, SHORTHAND_ICONS 
from pathlib import Path

import logging

logger = logging.getLogger(__name__)

__default_domain_actions: dict[Literal["domain"],Optional[Literal["action"]]] = {  
                        "default": None,
                        "switch": "switch.toggle",
                        "light": "light.toggle",
                        "button": "button.press",
                        "climate": "climate.toggle",
                        "fan": "fan.toggle",
                        "media_player": "media_player.toggle",
                        "remote": "remote.toggle",
                        "scene": "scene.turn_on",
                        "select": "select.next", ##For this one, cycle should be set to true automatically...
                        
                        "automation": "automation.trigger",
                        "script": "script.turn_on",
                        
                        "input_button": "input_button.press",
                        "input_boolean": "input_boolean.toggle",
                        "input_select": "input_select.select_next"
                        }

DEFAULT_DOMAIN_ACTIONS = __default_domain_actions
"""
Dict mapping certain HA domains to default actions that can be applied to elements without entity domain restrictions.
Domains without a default action are not present in this dict.
"""

all_entities_config = CORE.config.configuration.get("entities",{}).copy()
"All entitites as defined in the config file"

all_entities = {"sun.sun": {"entity_id": "sun.sun"}}
for entity_config in all_entities_config:
    if "entity_id" not in entity_config:
        logger.error(f"Entries in the entity config require an entity_id. Cannot add {entity_config}")
    else:
        all_entities[entity_config["entity_id"]] = entity_config

entity_tags : dict
"Entities defined in the entities.yaml file. Can be used to parse entities by prefixing the key with !entity (Same as in the yaml config)"

base_folder = CORE.config.folders["base_folder"]

entity_file = base_folder / "entitities.yaml"
if entity_file.exists():
    with open(entity_file) as f:
        entity_tags = yaml.load(f, Loader=yaml.SafeLoader)
    f.close()

    for tag, entity in entity_tags.items():
        if entity not in all_entities:
            all_entities[entity] = {"entity_id": entity}
else:
    entity_tags = {}

ENTITY_TAG_KEY = "!entity "
def parse_entity_tag(entity : str) -> Union[str,Literal[False]]:
    if entity.startswith(ENTITY_TAG_KEY):
        tag = entity.removeprefix(ENTITY_TAG_KEY)
        
        if tag not in entity_tags:
            msg = f"{tag} could not be found as a key in the entities.yaml file. "
            logger.exception(KeyError(msg))
            return False
        else:
            return entity_tags[tag]

_all_services_config = CORE.config.configuration.get("service_actions",{}).copy()

##Generally this should happen during import I think, and return a False maybe for the setup?
all_service_actions = {}
for service_config in _all_services_config:
    if "service_id" not in service_config:
        logger.error(f"Entries in the service_actions config require a service_id. Cannot add {service_config}")
    else:
        all_service_actions[service_config["service_id"]] = service_config

DEFAULT_HA_DT_FORMAT = "%Y-%m-%-dT%H:%M:%S.%Y+%H:%M"
"""
The default format Home Assistant seems to use for datetime strings.
Parsing this using `datetime.strptime()` throws an error, however it does work when using `datetime.fromisoformat()`
That means this constant is likely not necessary anywhere, but leaving it here in case.
"""

DEFAULT_PING_INTERVAL : int = 50 #seconds
"Default time in seconds to send a new ping"

MAX_PONGS_MISSED : int = 5
"Max amount of pongs to be missed before the connection is considered broken."

HOMEASSISTANT_BLUE : tuple = (3, 169, 244, 255)
"The Blue Color used in Home Assistant Branding :)"

HOMEASSISTANT_ICON = Path(__file__).parent / 'home-assistant.png'

ERROR_STATES = {"unknown", "unavailable"}
"Shorthand for the states that indicate an error in the entity (so unknown or unavailable)"

UNKNOWN_ICON = "mdi:help"
"Default icon for unknown states"

UNAVAILABLE_ICON = "mdi:exclamation-thick"
"Default icon for unavailable states"

UNAVAILABLE_COLOR = "gray4"
"Color to use for text elements when the entity state is unavailable"

UNKNOWN_COLOR = "gray4"
"Color to use for text elements when the entity state is unknown"


cf = CORE.config.configuration["home_assistant"]
if "unavailable_color" in cf:
    UNAVAILABLE_COLOR = cf["unavailable_color"]

if "unknown_color" in cf:
    UNKNOWN_COLOR = cf["unknown_color"]

if "unavailable_icon" in cf:
    UNAVAILABLE_ICON = cf["unavailable_icon"]

if "unknown_icon" in cf:
    UNKNOWN_ICON = cf["unknown_icon"]

if "ping_pong_interval" in cf:
    DEFAULT_PING_INTERVAL = cf["ping_pong_interval"]



from typing import *
import tkinter as tk

import ttkbootstrap as ttk

from inkBoard import core as CORE

from inkBoarddesigner.tkinter import window, functions as tk_functions
from inkBoarddesigner.tkinter.widgets import Treeview
from inkBoarddesigner import const as des_const

from .const import ENTITY_ICONS
from .. import async_setup as super_setup, async_start as super_start, logger as _LOGGER

if TYPE_CHECKING:
    from inkBoard import config
    from PythonScreenStackManager import PSSMScreen
    from .. import client

entity_tree= ttk.Treeview(columns=("entity"),
                                    name="entity-tree")

entity_tree = Treeview(entity_tree)
entity_tree.heading("#0", text="Entity", anchor="w")
entity_tree.heading("#1", text="State", anchor="w")

entity_tree.column("#0", minwidth=100, width=int(des_const.INTERFACE_WIDTH*0.6))
entity_tree.column("#1", minwidth=50, width=int(des_const.INTERFACE_WIDTH*0.4))

def get_client() -> "client.HAclient":
    return CORE.integration_objects[__package__.split(".")[-2]]


async def async_setup(core: "CORE", config : "config"):

    window.treeFrame.register_tree("Entities",entity_tree)
    return await super_setup(core, config)

async def async_start(core: "CORE", client : "client.HAclient"):
    client._subcribe_callbacks.append(_new_subscription)
    await super_start(core.screen, client)
    build_entity_tree(client)

    for entity in client.stateDict:    
        client.add_entity_function(
            entity, ((update_tree_row, False),)
        )
    return


def build_entity_tree(client:  "client.HAclient"):

    for id in client.stateDict:
        add_entity_to_tree(client, id)
    return


def select_tree(tree, event, iids):

    client: "client.HAclient" = CORE.integration_objects[__package__.split(".")[-2]]

    iid = iids[0]

    if iid in client.elementDict:
        elts = client.elementDict[iid]
        tk_functions.highlight_elements(*elts)
        return

    entity_tree.highlight_element(tree,event,iids)

def _new_subscription(client: "client.HAclient", header: dict):
        
    entity = header["trigger"].get("entity_id", None)
    if not entity: return

    add_entity_to_tree(client, entity)
    return


def add_entity_to_tree(client: "client.HAclient", entity: str):
    id = entity
    state = client.stateDict.get(id, None)
    
    if not (entity or state): 
        return

    open_init = False

    if not entity_tree.exists(id):
        if id in client.elementDict:
            idx = 0
        else:
            idx = tk.END

        dom = id.split(".")[0]
        icon_mdi = ENTITY_ICONS.get(dom,ENTITY_ICONS["default"])
        icon = tk_functions.get_tree_icon(icon_mdi)
        name = state["attributes"].get("friendly_name", id)
        entity_tree.insert(
            "",
            idx,
            iid = id,
            text=name,
            values=(state["state"]),
            image=icon,
            open=open_init
        )
    else:
        if id in client.elementDict:
            if not entity_tree.get_children(id):
                entity_tree.move(id,"",0)

    for elt in client.elementDict.get(id,{}):
        for child in entity_tree.get_children(id):
            if elt.id in child:
                break
        
        if [x for x in entity_tree.get_children(id) if elt.id in x]:
            _LOGGER.debug(f"Duplicate element to be added {elt.id}")
            break
        

        elttype = str(elt.__class__.__name__)
        entity_tree.insert_element_item(
            elt, elttype, id
            )   
    return

def update_tree_row(trigger_dict, state_dict):
    iid = trigger_dict["entity_id"]
    state = trigger_dict["to_state"]["state"]
    entity_tree.item(iid, values=(state))
    return

def show_entity_tip(tree: Treeview, event, _iid):

    client = get_client()
    state = client.stateDict.get(_iid, None)

    if not state: 
        tree.tooltip.hide_tip()
        return

    tipText = (f"entity id: {state['entity_id']}\n"
            f"state: {state['state']}")
    if "attributes" in state:
        tipText = tipText + "\nattributes:"
        attributes = state['attributes']
        for attr,val in sorted(attributes.items()):
            tipText = tipText + "\n" + "   " + f"{attr}: {val}"
    tree.tooltip.text = tipText
    tree.tooltip.show_tip()
    return

entity_tree.on_select = select_tree
entity_tree.on_hover = show_entity_tip


#!/usr/bin/env python

"inkBoard platform for kobo, and likely other devices with FBInk installed"

import sys
import os
import socket
import asyncio
import logging
import subprocess
import concurrent.futures
from typing import *

from pathlib import Path
from time import sleep
from math import pi, ceil
from contextlib import suppress
# Load the wrapper module, it's linked against FBInk, so the dynamic loader will take care of pulling in the actual FBInk library

#Fbink functions etc. can best be checked here: https://github.com/NiLuJe/FBInk/blob/master/fbink.h
#But this is all in C, so some translating may be needed -> yawk has an fbink mock
# Load Pillow

from PythonScreenStackManager import constants as const, devices as basedevice, tools
from PythonScreenStackManager.tools import DummyTask, TouchEvent
from PythonScreenStackManager.pssm_types import *
from PythonScreenStackManager.pssm.util import elementactionwrapper

import inkBoard.constants
from inkBoard.platforms.basedevice import BaseDevice, BaseConnectionNetwork, InkboardDeviceFeatures, FEATURES

from PIL import Image, ImageDraw, ImageFont, ImageOps

from . import aioKIP, util, pssm_device
from .aioKIP import InputQueue
from .fbink import API as FBInk

_LOGGER = logging.getLogger(__name__)


try:
	import pywifi ##Implement pywifi rn but for the bare pssm implementation don't implement it.
	pywifi_installed = True
	_LOGGER.debug("Using pywifi")
except ModuleNotFoundError:
	pywifi_installed = False

class Device(BaseDevice, pssm_device.Device):
	"""Base class for inkBoard on kobos.

	Optionally can control the wifi interface.
	"""
	def __init__(self, name: str = pssm_device.full_device_name, rotation: RotationValues = "UR", kill_os: bool = True, refresh_rate: DurationType = "30min",
			touch_debounce_time: DurationType = aioKIP.DEFAULT_DEBOUNCE_TIME, hold_touch_time: DurationType = aioKIP.DEFAULT_HOLD_TIME, input_device_path: str = aioKIP.DEFAULT_INPUT_DEVICE):
		
		features = pssm_device.feature_list.copy()
		if pywifi_installed:
			_LOGGER.debug("Setting up connection Network")
			self._network = ConnectionNetwork()
			features.append(FEATURES.FEATURE_CONNECTION)
		else:
			_LOGGER.debug("Setting up Base Network")
			self._network = pssm_device.Network()

		##Check if these are correctly parsed when calling event bindings
		self.__KIPargs = {"input_device": input_device_path}
		self.__KIPargs["debounce_time"] = tools.parse_duration_string(touch_debounce_time)
		self.__KIPargs["long_click_time"] = tools.parse_duration_string(hold_touch_time)

		self._model = pssm_device.full_device_name
		self._name = name
		self._features = InkboardDeviceFeatures(*features)

		self._battery = pssm_device.Battery()
		self._backlight = pssm_device.Backlight(self)

		if kill_os:
			util.kill_os()
		
		tools.parse_duration_string(refresh_rate)
		self._refreshRate = refresh_rate

		FBInk.rotate_screen(rotation)
		splashscreen = inkBoard.constants.INKBOARD_FOLDER / "files" / "images" / "default_background.png"
		splash_img = ImageOps.fit(Image.open(splashscreen),(self.screenWidth,self.screenHeight))
		FBInk.fbink_print_pil(splash_img)

	#region
	##Redefining a few properties to prevent having to call the basedevice
	@property
	def colorType(self) -> Image.ImageMode:
		"Same as screenMode. Implemented for legacy purposes"
		return  self.screenMode
	
	@property
	def screenMode(self) -> Image.ImageMode:
		"The mode of images being printed on the screen"
		return "LA"
	
	@property
	def imgMode(self):
		return "RGBA"

	@property
	def screenType(self):
		return "E-Ink"
	
	@property
	def refreshRate(self) -> DurationType:
		"The interval between which the screen is fully refreshed"
		return self._refreshRate
	#endregion

	def print_pil(self, imgData, x, y, isInverted=False):
		_LOGGER.debug("Printing to device screen")
		if imgData.mode != self.screenMode:
			imgData = imgData.convert(self.screenMode)

		FBInk.fbink_print_pil(imgData,x,y)
		
	async def async_pol_features(self):
		await self.battery.async_update_battery_state()
		await self.network.async_update_network_properties()
		return

	async def event_bindings(self, touch_queue = None):
		asyncio.create_task(self.refresh_loop())
		self._eventQueue = InputQueue(**self.__KIPargs)
		with suppress(asyncio.CancelledError):
			while self.Screen.printing:
				(x,y,action) = await self.eventQueue.get()
				if action == aioKIP.TOUCH_SHORT:
					touch_action = const.TOUCH_TAP
				else:
					touch_action = const.TOUCH_LONG

				await touch_queue.put(TouchEvent(x,y,touch_action))
		return

	async def refresh_loop(self):
		wait_time = tools.parse_duration_string(self.refreshRate)
		while self.Screen.printing:
			try:
				await asyncio.sleep(wait_time)
				self.refresh_screen()
			except asyncio.CancelledError:
				return
		
	async def _rotate(self, rotation=None):
		await asyncio.to_thread(FBInk.rotate_screen(rotation))
		await asyncio.to_thread(FBInk.screen_refresh())

	@elementactionwrapper.method
	def clear_screen(self):
		"Clears the entire screen"
		FBInk.screen_clear()
	
	@elementactionwrapper.method
	def refresh_screen(self, skip_clear: bool = False):
		"Refreshes the entire screen. By default clears it first"
		if not skip_clear:
			self.clear_screen()
		
		FBInk.screen_refresh()

	def set_waveform(self, mode):
		FBInk.set_waveform(mode)
	
class ConnectionNetwork(pssm_device.Network, BaseConnectionNetwork):
	def __init__(self):
		wifilogger = logging.getLogger(pywifi.__name__)
		wifilogger.setLevel(logging.WARNING)
		wifi = pywifi.PyWiFi()
		self._iface: pywifi.iface.Interface = wifi.interfaces()[0]
		profiles = self._iface.network_profiles()
		if profiles:
			self._baseprofile: pywifi.Profile = profiles[0]
		else:
			self._baseprofile = None
		super().__init__()

	def get_network_properties(self):
		##Add function to get network name

		if s := self._iface.status() == pywifi.const.IFACE_INACTIVE:
			self._isWifiOn = False
			self._connected = False
		elif s == pywifi.const.IFACE_CONNECTED:
			self._isWifiOn = True
			self._connected = True
		else:
			self._isWifiOn = True
			self._connected = False

		self._macAddr = util.get_mac()
		if self.connected:
			self._IP = util.get_ip()
			self._SSID = util.get_SSID()
		else:
			self._IP = None
			self._SSID = None

	async def async_connect(self, ssid: str = None, password: str = None):
		"""Connects to the wifi"""
		if ssid == None:
			profile = self._baseprofile
		else:
			_LOGGER.warning("Connecting to custom networks may not work")
			profile = pywifi.Profile()
			profile.ssid = ssid
			profile.auth = pywifi.const.AUTH_ALG_OPEN
			profile.akm.append(pywifi.const.AKM_TYPE_WPA2PSK)
			profile.cipher = pywifi.const.CIPHER_TYPE_CCMP
			profile.key = password
		
		await asyncio.to_thread(self.__wifi_connect, profile)

	def connect(self, ssid: str = None, password: str = None):
		asyncio.create_task(self.async_connect(ssid,password))

	def __wifi_connect(self, profile: "pywifi.Profile"):

		if profile != self._baseprofile:
			self._baseprofile = profile

		self._iface.connect(profile)
		return

	async def async_disconnect(self):
		await asyncio.to_thread(self.__wifi_disconnect)

	def disconnect(self):
		asyncio.create_task(self.async_disconnect())

	def __wifi_disconnect(self):
		self._iface.disconnect()




from inkBoard import core as CORE

from . import HAelements, clientelements

d_HA = CORE.util.get_module_elements(HAelements)
d_client = CORE.util.get_module_elements(clientelements)

element_dict = d_HA | d_client #dict with all elements of the Home Assistant integration

def parse_ha_element(elt_type : str):
    return element_dict.get(elt_type,None)


# CORE.add_element_parser("HA",parse_ha_element)

def setup_elements():
    clientelements.HomeAssistantMenu()
    clientelements.menu.StatusBar.add_statusbar_element("home-assistant",clientelements.ClientElement())




import logging
from typing import Literal
from math import floor
import tkinter as tk


import ttkbootstrap as ttk
from ttkbootstrap.tooltip import ToolTip

from mdi_pil.ttkbootstrap_mdi import MDIIcon, MDIButton

from .. import const
from ..settings import EM_SETTINGS

from . import window
from .windows import DesignerWindow, TreeFrame
from .widgets import LabelToggle, LabelIcon, Treeview
from . import functions as tk_functions

logger = logging.getLogger(__name__)

_KEEP = []


def build_tk_button(root: tk.Tk, button_size : tuple[Literal["w"],Literal["h"]], icon : str, text : str, command=None, widget_name:str = None) -> tk.Button:
    """
    Builds a button for the tk window using the default values. Returns the button and image source (since that needs to be saved somewhere too.)
    Also makes the image to put upon it in the correct font.
    args:
        root: tk instance
        button_size: size of the button as a (w,h) tuple
        icon: mdi icon string
        text: text to show on the button
        command: optional command to call when the button is clicked
    """
    
    textSize = const.BUTTON_FONT_SIZE
    iconSize = int(textSize*1.5)

    imgTk = MDIButton(icon, text,button_size,const.HA_FONT_FILE,
                        iconSize,textSize, icon_margin=(0,3)
                    )
    _KEEP.append(imgTk)

    button = ttk.Button(root,image=imgTk,command=command, cursor=const.INTERACT_CURSOR, style=const.BUTTON_STYLE, padding=0,
                        takefocus=False, name=widget_name)
    return button

def build_interface_buttons(interface_frame: ttk.Frame):
    "builds the buttons used to run various designer functions"

    buttFrame = ttk.Frame(interface_frame)
    buttFrame.grid(row=0, column=0, pady=5, sticky=tk.NSEW)

    labelFrame = ttk.Frame(buttFrame)
    labelFrame.grid(row=0, column=0, columnspan=2, sticky=tk.NSEW, padx=const.SETTINGS_PADDING)

    baselabel = ttk.Label(labelFrame, text="inkBoard", cursor="",
            font=const.DEFAULT_LABEL_FONT, anchor=tk.W,
            style="primary.TLabel")
    baselabel.pack(side=tk.LEFT)

    label = ttk.Label(labelFrame, text="Config", cursor="",
                    font=const.DEFAULT_LABEL_FONT, anchor=tk.W, 
                    bootstyle=(ttk.PRIMARY),  style="TLabel")
    label.bind("<Button-1>", tk_functions.open_config_folder)
    ToolTip(label, const.CONFIG_LABEL_TIP, bootstyle=const.TOOLTIP_STYLE)

    open_icon = MDIIcon("mdi:folder-open", (15,15))
    _KEEP.append(open_icon)

    open_button = ttk.Button(labelFrame,image=open_icon, 
                            style='image.TButton', 
                            cursor=const.INTERACT_CURSOR, command=tk_functions.open_new_config)
    ToolTip(open_button,const.CONFIG_OPEN_TIP, bootstyle=const.TOOLTIP_STYLE)
    
    open_button.pack(side=tk.RIGHT)

    window._buttonFrame = buttFrame
    window._buttonLabel = baselabel
    window._configLabel = label

    screenshotButton = build_tk_button(buttFrame,
                                    (const.BUTTON_WIDTH,const.BUTTON_HEIGHT),const.SCREENSHOT_BUTTON_ICON,const.SCREENSHOT_BUTTON_TEXT, 
                                    command=tk_functions.make_screenshot, widget_name=const.SCREENSHOT_BUTTON_NAME)
    screenshotButton.grid(column=0, row=1, 
                        sticky=tk.NW,padx=const.SETTINGS_PADDING)
    ToolTip(screenshotButton,const.SCREENSHOT_TIP, bootstyle=const.TOOLTIP_STYLE)

    packButton = build_tk_button(buttFrame,(const.BUTTON_WIDTH,const.BUTTON_HEIGHT),
                                const.PACK_BUTTON_ICON,const.PACK_BUTTON_TEXT, 
                                command=tk_functions.make_package, widget_name=const.PACK_BUTTON_NAME)
    packButton.grid(column=0, row=2, 
                    sticky=tk.NW, padx=const.SETTINGS_PADDING)
    ToolTip(packButton,const.PACK_TIP, bootstyle=const.TOOLTIP_STYLE)

    reloadButton = build_tk_button(buttFrame,(const.BUTTON_WIDTH,const.BUTTON_HEIGHT),
                                        const.RELOAD_BUTTON_ICON,const.RELOAD_BUTTON_TEXT, 
                                        command=tk_functions.reload_config, widget_name=const.RELOAD_BUTTON_NAME)
    reloadButton.grid(column=1, row=1, 
                    sticky=tk.NE, padx=(0,const.SETTINGS_PADDING))
    ToolTip(reloadButton,const.RELOAD_TIP, bootstyle=const.TOOLTIP_STYLE)

    quitButton = build_tk_button(buttFrame,(const.BUTTON_WIDTH,const.BUTTON_HEIGHT),
                                        const.STOP_BUTTON_ICON, const.STOP_BUTTON_TEXT, 
                                        tk_functions.stop_emulator, const.STOP_BUTTON_NAME)
    quitButton.grid(column=1, row=2, 
                    sticky=tk.NE, padx=(0,const.SETTINGS_PADDING))
    ToolTip(quitButton,const.STOP_TIP, bootstyle=const.TOOLTIP_STYLE)

    window.interface_buttons = (screenshotButton, packButton, reloadButton, quitButton)


def build_label_fill(root, size : tuple[Literal["w"],Literal["h"]], text:str, tiptext:str, var : tuple[tk.Variable,"function"]=(None,None), command=None):
    w,h = size

    frame = LabelToggle(root,text=text,
                        width=w,height=h,
                        style='noframe.TLabelframe', 
                        labelanchor="w", padding=10, cursor=const.INTERACT_CURSOR)
    ToolTip(frame, text=tiptext,bootstyle=const.TOOLTIP_STYLE)
    return frame

def build_label_toggle(root, size : tuple[Literal["w"],Literal["h"]], text:str, tiptext:str, var: tk.Variable = None, command=None):
    "Builds toggle buttons with labels and a tooltip"
    w,h = size
    frame = LabelToggle(root,text=text,
                        width=w,height=h, variable=var, command=command,
                        labelanchor="w", padding=10, cursor=const.INTERACT_CURSOR)
    ToolTip(frame, text=tiptext,bootstyle=const.TOOLTIP_STYLE)
    return frame

def build_label_icon(root,size : tuple[Literal["w"],Literal["h"]], text:str, tiptext:str, var : tuple[tk.Variable,"function"]=(None,None), command=""):
    w,h = size
    icon = "mdi:cog"
    icon_size = int(h*0.75)
    frame = LabelIcon(icon, icon_size, text=text,
                    master=root, width=w,height=h,padding=10, cursor=const.INTERACT_CURSOR, command=command)
    ToolTip(frame, text=tiptext,bootstyle=const.TOOLTIP_STYLE)

    return frame

def build_settings_frame(interface_frame: ttk.Frame):
    "Builds the frames that holds the settable option widgets"

    settFrame = ttk.Labelframe(interface_frame, text="Settings",
                            bootstyle=ttk.PRIMARY,
                            padding=0)
    settFrame.grid(column=0, row=1, sticky="nsew", padx=0, pady=(5,0))

    window._settingsFrame = settFrame

    dmVar = window.darkmode_variable
    dmFrame = build_label_toggle(settFrame,(const.SETTINGS_WIDTH, int(const.SETTINGS_HEIGHT/2)),
                                        text="Dark Mode", tiptext=const.DARKMODE_TIP,command=tk_functions.change_theme, var=dmVar)
    dmFrame.grid(row=0,column=0, sticky="e",padx=const.SETTINGS_PADDING)

    saveVar = window.saveas_variable
    saveFrame = build_label_toggle(settFrame,(const.SETTINGS_WIDTH, int(const.SETTINGS_HEIGHT/2)), 
                        text="Save As", tiptext=const.SAVEAS_TIP, var=saveVar)
    saveFrame.grid(row=2,column=0, sticky="e",padx=const.SETTINGS_PADDING)

    hlVar = window.hightlight_variable
    hlFrame = build_label_toggle(settFrame,(const.SETTINGS_WIDTH, int(const.SETTINGS_HEIGHT/2)), 
                        text="Highlight", tiptext=const.HIGHLIGHT_TIP, var=hlVar)
    hlFrame.grid(row=1,column=0, sticky="e",padx=const.SETTINGS_PADDING)

    blVar = tk.BooleanVar(window, value=EM_SETTINGS.getboolean("backlight"), name="backlight")
    blFrame = build_label_icon(settFrame,(const.SETTINGS_WIDTH, int(const.SETTINGS_HEIGHT/2)), 
                        text="Config", tiptext=const.CONFIG_OPTIONS_TIP, command=tk_functions.open_config_window)
    blFrame.grid(row=0,column=1, sticky="e",padx=const.SETTINGS_PADDING)
    
    deviceFrame = build_label_icon(settFrame,(const.SETTINGS_WIDTH, int(const.SETTINGS_HEIGHT/2)), 
                        text="Device", tiptext=const.DEVICE_TIP, command=tk_functions.open_device_window)
    deviceFrame.grid(row=1,column=1, sticky="e",padx=const.SETTINGS_PADDING)

    listFrame = ttk.Labelframe(settFrame,text="List",
                            height=int(const.SETTINGS_HEIGHT/2), style='noframe.TLabelframe', labelanchor="w")
    listFrame.grid(row=3,column=0, columnspan=2, sticky="ns", pady=0, padx=const.SETTINGS_PADDING)

    listFrame.grid_columnconfigure(0, weight=7)
    listFrame.grid_columnconfigure(1, weight=3)

    menu = ttk.Combobox(listFrame, state="readonly", values = const.DEFAULT_LIST_OPTIONS, 
                        width=const.LIST_WIDTH, textvariable=window._tree_list_variable)
    window._list_menu = menu

    menu.grid(row=0, column=0, sticky="NSEW")

    icon_size = int((const.SETTINGS_HEIGHT/2)*0.6)
    imgTk = MDIIcon("mdi:refresh", (icon_size,icon_size))
    _KEEP.append(imgTk)

    button = ttk.Button(listFrame, image=imgTk, cursor="hand2", padding=int(icon_size/4), style="image.TButton")
    ToolTip(button,const.TREE_REFRESH_TIP, bootstyle=const.TOOLTIP_STYLE)
    button.grid(row=0,column=1, sticky="nsew")

def build_tree_frame(frame: ttk.Frame):

    treeFrame = TreeFrame(frame, padding=0, name=const.TREE_FRAME_NAME)
    treeFrame.pack(anchor=tk.S, fill=tk.BOTH, expand=1, ipady=10)

    treeFrame._window = window
    window._treeFrame = treeFrame

    element_treeview = Treeview(ttk.Treeview(columns=("entity"),))
    element_treeview.heading("#0", text="Element", anchor="w")
    element_treeview.column("#0", width= floor(const.INTERFACE_WIDTH*0.375)) ##Width and height are empty
    element_treeview.heading("entity", text="Entity", anchor="w")
    element_treeview.column("entity", width= floor(const.INTERFACE_WIDTH*0.475))
    
    treeFrame._base_trees[const.ELEMENT_TREE_OPTION.lower()] = element_treeview
    treeFrame._element_tree = element_treeview
    return


def build_interface():
    "Builds the button interface for the designer"

    ui_frame = ttk.Frame(window, name=const.UI_FRAME_NAME,
                            width=const.INTERFACE_WIDTH-2*const.INTERFACE_BORDER, borderwidth=const.INTERFACE_BORDER,
                            style='right.TFrame', 
                            )

    iface_frame = ttk.Frame(ui_frame,
                            style='buttons.TFrame')
    iface_frame.pack()

    build_interface_buttons(iface_frame)
    build_settings_frame(iface_frame)
    build_tree_frame(ui_frame)
    return ui_frame


def create_canvas(size: tuple[int,int]):
    "Creates the canvas widget that the pssm screen will print on"

    canvas = window.setup_canvas(size)    
    return canvas

def build_window() -> DesignerWindow:

    title = "inkBoard Designer"
    window.wm_title(title)
    
    init_width = const.INTERFACE_WIDTH*4
    init_height = int((init_width*9)/16)

    window.geometry(f"{init_width}x{init_height}")
    window.focus_force()

    iface = build_interface()
    window.setup_canvas((init_width-const.INTERFACE_WIDTH, init_height))

    iface.place(anchor=tk.NE,width=const.INTERFACE_WIDTH,relx=1,rely=0, relheight=1)
    return window




#!/usr/bin/env pythonall_entities

import asyncio
import json
from datetime import datetime
import logging
from typing import Union, Callable, TYPE_CHECKING, TypedDict, Optional, Literal, TypeVar, Any, Coroutine, Sequence
from types import MappingProxyType

import websockets
from websockets import protocol as ws_protocol
from websockets.asyncio import client as ws_client

from PythonScreenStackManager.pssm import screen

import inkBoard
from inkBoard.constants import FuncExceptions
from inkBoard.platforms import FEATURES

from PythonScreenStackManager import tools, elements
from PythonScreenStackManager.tools import DummyTask

from inkBoard import core as CORE

from .helpers import triggerDictType, stateDictType, actionCallDict, EntityType
from .constants import all_entities, all_service_actions, entity_tags, ENTITY_TAG_KEY, parse_entity_tag, \
                        DEFAULT_PING_INTERVAL, MAX_PONGS_MISSED, DEFAULT_HA_DT_FORMAT



from .HAelements import HAelement
from .clientelements import ClientElement

if TYPE_CHECKING:
    from PythonScreenStackManager.devices import PSSMdevice as pssm_device
    from PythonScreenStackManager import elements, tools #import pssm as screenFile, elements as pssm
    from PythonScreenStackManager.tools import DummyTask

_LOGGER = logging.getLogger(__name__)
_LOGGER.debug(f"{_LOGGER.name} has loglevel {logging.getLevelName(_LOGGER.getEffectiveLevel())}")

class ServerConfigDict(TypedDict):
    "The entries saved from the servers config response"

    name : str
    "The name of the instance"
    
    time_zone : str
    "The server's time zone"

    version : str
    "The version of Home Assistant that the server is running"

    integration : bool
    "Boolean value indicating if the inkBoard integration is running on the server"

    unit_system : dict
    "The default units this server uses"

#Builds the headers to send to subscribe to triggers
def trigger_headers(entities : Union[list,tuple],last_id : int):
    """
    Builds a list of headers to send to a client to subscribe to state changes of the entities in entities
        Parameters:
            entities: iterable with strings containing the entity_id's to subscribe to
            last_id: the last message id send to the websocket API
    """
    headers = []
    entity_list = []
    if type(entities) is str:
        #Simple rewite to keep the code below functioning in case of inputting a single entity
        entities = [entities]

    for index, entity in enumerate(entities):
        subscribe_header = {
            "id": last_id + index+1,
            "type": "subscribe_trigger",
            "trigger": {
                "platform": "state",
                "entity_id": entity
            }, }
        headers.append(subscribe_header)
        entity_list.append((entity))
    return headers

class HAclient:
    '''
    Client class that manages the connection with Home Assistant. Subscribes to triggers and updates elements.
        screen: instance of PSSMScreen that handles the printing of layouts and elements etc
    '''
    
    def __init__(self, screen: screen.PSSMScreen, ping_interval:int = DEFAULT_PING_INTERVAL):

        self._websocket: ws_client.ClientConnection
        self._pssmScreen = screen
        self.pssmScreen.client = self

        self._subcribe_callbacks = [] 
        #Callbacks to call when subscribing to a new entity. None async functions only -> eh no make tasks of them and then continue.

        screen.add_shorthand_function("service-action",self.call_service_action)
        screen.add_register_callback(self.register_new_element)
        
        self.hass_data = CORE.config.configuration["home_assistant"]
        self.__last_id = 0
        self.loglist = []
        self._HAconfig = ServerConfigDict(name=None, time_zone=None, version=None, integration=False, unit_system={})
        self._elementDict : dict[str,set[EntityType]] = {}
        self._functionDict : dict[str,list] = {}
        self._stateDict : dict = {}
        self.updatingAll : bool = False

        self.__websocketCondition = asyncio.Condition()

        self.ping_interval = ping_interval
        self.listenerTask : asyncio.Task = DummyTask()
        self.commanderTask : asyncio.Task = DummyTask()
        self.pingpongTask : asyncio.Task = DummyTask()
        self.longrunningTasks: asyncio.Task = DummyTask()
        self.reconnect_task : asyncio.Task = DummyTask()

        self._callback_queues : dict["id", asyncio.Queue]= {} ##Dict with keys corresponding to message id's, values being asyncio events
        self.__message_queue = asyncio.Queue()
        self._commanderLock = asyncio.Lock()
        self._listenerLock = asyncio.Lock()

    #region client properties
    @property
    def websocket(self) -> ws_client.ClientConnection:
        "The websocket client object managing the connection."
        if hasattr(self,"_websocket"):
            return self._websocket
        else:
            return None
    
    @property
    def HAconfig(self) -> dict:
        "The configuration from the Home Assistant server (only keys needed for inkboard)"
        return self._HAconfig
    
    @property
    def next_id(self) -> int:
        "Returns the next message id and increments last_id"
        self.__last_id += 1
        return self.__last_id

    @property
    def connection(self) -> bool:
        "True if the client websocket is open."
        if self.websocket == None:
            return False
        else:
            return self.websocket.state is ws_protocol.State.OPEN
    
    @property
    def clientState(self) -> Literal["disconnected", "connecting", "connected"]:
        "Returns the state the client is currently in"
        if self.websocket == None or self.connectionTask.done():
            return "disconnected"
        else:
            if self.connection:
                return "connected"
            else:
                return "connecting"

    @property
    def websocketCondition(self) -> asyncio.Condition:
        "Condition that is notified when something happends to the websocket connection, generally in case of loss of connect, or start of a reconnect."
        return self.__websocketCondition

    @property
    def pssmScreen(self) -> screen.PSSMScreen:
        "The PSSM screen object the client is managing"
        return self._pssmScreen

    @property
    def device(self) -> "pssm_device":
        "The device object connected to the PSSM screen object"
        return self._pssmScreen.device
    
    @property
    def elementDict(self) -> dict[EntityType, HAelement]:
        "Dict that links entities to pssm elements that need to be updated with the entity"
        return MappingProxyType(self._elementDict)
    
    @property
    def functionDict(self) -> MappingProxyType[EntityType, Callable[[triggerDictType,"HAclient"],Any]]:
        "Dict that links entity id's to functions that have to be called when its state update"
        return MappingProxyType(self._functionDict)
    
    @property
    def stateDict(self)-> MappingProxyType[EntityType, stateDictType]:
        "dict with the current states of all subscribed to entities"
        return MappingProxyType(self._stateDict)
    
    @property
    def messageQueue(self) -> asyncio.Queue:
        "Queue with messages to send to the server"
        return self.__message_queue
    
    @property
    def commanding(self) -> bool:
        "True if the client is currently able to send commands to the Home Assistant Server (e.g. call service actions etc.)."
        return self._commanderLock.locked()
    
    @property
    def listening(self) -> bool:
        "True if the client is currently listening to the Home Assistant Server"
        return self._listenerLock.locked()
    
    @property
    def server_config(self) -> ServerConfigDict:
        "The important parts of the configuration of the Home Assistant server"
        return MappingProxyType(self._HAconfig)

    async def await_commander(self) -> None:
        """
        This can be used to await for the client to start accepting commands, i.e. know when it's possible to call services or call other stuff on the websocket.
        """
        async with self._commanderLock:
            await asyncio.sleep(0)
    #endregion

    #region [websocket stuff]
    def reconnect_client(self, initWait=5, *args, **kwargs):
        """Starts the task to reconnect to the client, and periodically retry doing so."""
        self.reconnect_task = self.loop.create_task(self.__async__reconnect(initWait))
    
    async def connect_client(self):
        """Starts the function that connects to Home Assistant""" 
        self.loop = asyncio.get_event_loop()
        self.connectionTask = asyncio.create_task(self.__async__connect())
        async with self.websocketCondition:
            await self.websocketCondition.wait()
        
        return

    async def disconnect_client(self):
        "Disconnects from the Home Assistant client."
        await self.websocket.close()
        async with self.websocketCondition:
            self.websocketCondition.notify_all()

    async def __async__connect(self):
        """Sets up a websocket connection to Home Assistant"""

        ##A rewrite of this could work by using a context manager to connect, and notify the condition when it is connected (which would then return the integration start function)
        ##And then use a gather call for the listener and commander
        ##Pros of this is that the context manager can automatically reconnect, and closes the connection when it exits too.
        ##Only thing that's kinda iffy: how would it signal when it is reconnecting?
        if not self.device.has_feature(FEATURES.FEATURE_NETWORK):
            _LOGGER.error("A device needs to have a network to use the Home Assistant integration")
            return

        uri = f"ws://{self.hass_data['url']}/api/websocket"
        token = self.hass_data["token"]
        auth_header =    {
            "type": "auth",
            "access_token": token     }
        _LOGGER.debug("Attempting connection to {}".format(uri))        

        async for websocket in ws_client.connect(uri, additional_headers=auth_header):
            _LOGGER.debug("Setting up websocket connection to Home Assistant")  
            try:                
                self._websocket = websocket
                await self.websocket.recv() #The first message send by the server requests authentication. Needs to be received to start it.
                await self.websocket.send(json.dumps(auth_header))
                auth_res = json.loads(await self.websocket.recv())
                if auth_res["type"] == "auth_ok":
                    self.authenthicated = True
                    _LOGGER.info(f"Connected to Home Assistant {auth_res}")
                else:
                    _LOGGER.error(f"Authentication failed {auth_res}")
                    return
                
                HAconf_header = {"id": self.next_id, "type": "get_config" }
                await self.websocket.send(json.dumps(HAconf_header))
                HAconf_res = json.loads(await self.websocket.recv())
                if HAconf_res["success"]:
                    HAconf_res = HAconf_res["result"]
                    self._HAconfig = {
                                    "name": HAconf_res["location_name"],
                                    "time_zone": HAconf_res["time_zone"],
                                    "version": HAconf_res["version"],
                                    "integration": "inkboard" in HAconf_res["components"], 
                                    "unit_system": HAconf_res["unit_system"]
                                    }

                states_header = {"id": self.next_id, "type": "get_states" } 
                await self.websocket.send(json.dumps(states_header))
                all_states = json.loads(await self.websocket.recv())
                if not all_states.get("success",False):
                    _LOGGER.error(f"Failed to get states, Home Assistant returned response: {all_states}")
                    continue
                else:
                    _LOGGER.debug("Received all states from Home Assistant")
                    initial_states = list(filter(lambda entity: entity["entity_id"] in all_entities, all_states["result"])) 
                    initial_dict = {}
                    for entity in initial_states:
                        initial_dict[entity["entity_id"]] = entity
                    self._stateDict = initial_dict
                    
                    timeout = CORE.config.inkBoard.integration_start_time
                    if isinstance(timeout, str):
                        timeout = tools.parse_duration_string(timeout)
                    if timeout < 0:
                        timeout = 5
                    
                    _LOGGER.debug("Updating all elements after connecting")
                    coro_list = []
                    coro_list.append(self.client_update_elements(update_all=True, timeout=timeout))
                    called_functions = [self.client_update_elements]
                    
                    _LOGGER.debug(f"Updating functions in function dict") #{self.functionDict}")
                    for func_entity in self.functionDict:
                        trigger_dict = triggerDictType(entity_id=func_entity, to_state=self.stateDict[func_entity], from_state=None, context=None)
                        for func_tuple in self.functionDict[func_entity]:
                            (func, call_func) = func_tuple
                            if func not in called_functions and call_func:
                                try:
                                    if asyncio.iscoroutine(func):
                                        coro_list.append(func(trigger_dict, self))
                                    else:
                                        coro_list.append(asyncio.to_thread(
                                            func, trigger_dict, self))
                                    called_functions.append(func)
                                except (TypeError, KeyError, IndexError, OSError) as exce:
                                    _LOGGER.warning(f"Error calling function {func} for entity {func_entity}: {exce}, removed from function dict")
                                    self._functionDict[func_entity].remove(func_tuple)
                    if coro_list:
                        done, pending = await asyncio.wait(coro_list,timeout=timeout)

                        for task in pending:
                            task : asyncio.Task
                            coro = task.get_coro()
                            _LOGGER.warning(f"{coro.__qualname__} is taking longer than the specified {timeout} while connecting, continuing in background and progressing connect script.")
                        
                        for task in done:
                            task : asyncio.Task
                            if task.exception() != None:
                                coro = task.get_coro()
                                _LOGGER.warning(f"{coro.__qualname__} raised an error while connecting: {task.exception()}")

                subscribe_headers = trigger_headers(all_entities,self.__last_id)
                subscribe_fails = 0

                for header in subscribe_headers:
                    res = await self.subscribe_to_trigger(None,header)
                    
                    if res == header:
                        continue
                    else:
                        subscribe_fails += 1
                if subscribe_fails == 0:
                    _LOGGER.info("Succesfully subscribed to all entities")
                
                self.__last_id = header["id"]
                async with self.websocketCondition:
                    self.websocketCondition.notify_all()

                if not self.longrunningTasks.done():
                    self.longrunningTasks.cancel()

                self.listenerTask = self.loop.create_task(self.__async_listen())
                self.commanderTask = self.loop.create_task(self.__async_command())
                runners = [self.listenerTask, self.commanderTask]

                self._longrunningTasks = asyncio.gather(*runners, return_exceptions=True)

                await self._longrunningTasks

                _LOGGER.debug("Listener and Commander task have returned")

            except websockets.exceptions.ConnectionClosed:
                if not self.longrunningTasks.done(): 
                    self.longrunningTasks.cancel("Client connection closed")
                _LOGGER.warning("Websocket connection has closed")
                continue
            except ConnectionRefusedError as exce:
                if not self.longrunningTasks.done(): self.longrunningTasks.cancel("Client connection closed")
                _LOGGER.error("Hme Assistant refused connection", exc_info=True)
            except asyncio.CancelledError:
                _LOGGER.debug("Connect task has been cancelled")
                if not self.longrunningTasks.done(): 
                    self.longrunningTasks.cancel("Client connection closed")
                return
            except Exception as e:
                _LOGGER.error(f"Something went wrong in the client: {e}")
                raise
        return

    async def __async__reconnect(self, init_Wait: float = 15, max_Attempts=0, wait_Increase: int =2, wait_Max: float = 300):
        """
        Reconnect function
        Args:
            init_wait: initial wait time (seconds) before reconnecting
            max_attempts: maximum attempts to try reconnecting before stopping the function
            wait_increase: how much the previous wait time is multiplied with for the next time; if init_wait = 0, next wait time will be 30 seconds
            wait_max: maximum waiting time (seconds) before attempting a new reconnection
        """
        wait = init_Wait
        att = 0
        _LOGGER.info("Reconnecting to Home Assistant")

        if self.connection:
            _LOGGER.info("Home Assistant is still connected, closing connection.")
            await self.disconnect_client()

        if not self.connectionTask.done():
            self.connectionTask.cancel("Attempting reconnect script")

        async with self.websocketCondition:
            self.websocketCondition.notify_all()

        while not self.connection:
            _LOGGER.debug("Trying to get IP")
            ip = await self.device.network.get_ip_async()
            if ip == None:
                _LOGGER.info("Got no IP adress, Trying to get SSID")
                ssid = await self.device.network.get_SSID_async()
                if ssid == "Wifi off":
                    _LOGGER.info("Wifi is still off. calling wifi up")
                    await self.device.network.wifi_up_async()
                else:
                    _LOGGER.info(f"Got network {ssid}")

                ip = await self.device.network.get_ip_async()
                if ip == None:
                    _LOGGER.error("Still got no IP adress, stopping reconnect")
                    return
            else:
                _LOGGER.debug(f"Got IP adress {ip}")

            await asyncio.sleep(wait)

            _LOGGER.info("Attempting reconnect")
            await self.connect_client()

            if not self.connection:
                if init_Wait == 0: init_Wait=30/2 #Otherwise it would stay at 30
                wait = init_Wait*(wait_Increase**(att + 1)) if init_Wait*(wait_Increase**(att + 1)) < 300 else 300
                att += 1
                _LOGGER.warning("Failed to reconnect, trying again in {} seconds".format(wait))

            if max_Attempts > 0 and att >= max_Attempts:
                _LOGGER.warning(f"Failed to reconnect within maximum attempts set: {max_Attempts}, stopping reconnect.")
                break

        async with self.websocketCondition:
            self.websocketCondition.notify_all()

        if self.connection:
            _LOGGER.info("Reconnected to Home Assistant")
        else:
            _LOGGER.warning("Failed to reconnect to Home Assistant")

    async def __async_listen(self):
        '''
        Listens to messages from the Home Assistant server
        '''
        _LOGGER.debug("Starting Listener")
        async with self._listenerLock:
            try:
                async for message in self.websocket: #@IgnoreException
                    message = json.loads(message)
                    id = message["id"]
                    _LOGGER.debug(f"Received message {id}")
                    _LOGGER.verbose(message)
                    if id in self._callback_queues:
                        queue = self._callback_queues[id]
                        _LOGGER.debug(f"Putting response from message id {id} in queue")
                        try:
                            queue.put_nowait(message)
                        except asyncio.QueueFull:
                            _LOGGER.error(f"Message id {id} has already queued a response")
                    
                    if message.get("type") == "event":
                        try:
                            asyncio.create_task(self.update_states(message))
                        except (TypeError, KeyError, IndexError, OSError) as exce:
                            _LOGGER.error(f"Error in update states for {message}: {exce}")
                    elif message.get("type") == "result":
                        if not message.get("success", False):
                            err = message.get("error",{})
                            _LOGGER.warning(f"Unsuccesful request {err.get('code','unknown code')}: {err.get('message','No message')}")
            except websockets.exceptions.ConnectionClosedError as exce:
                _LOGGER.error(f"Listener stopped due to connection closing")
                _LOGGER.debug(exce)
            except asyncio.CancelledError:
                pass
                        
        _LOGGER.warning("Listener stopped")
        if not self.commanderTask.done():
            self.commanderTask.cancel()
        async with self.websocketCondition:
            self.websocketCondition.notify_all()

    async def __async_command(self):
        '''
        Sends messages put into HAclient.messageQueue to the Home Assistant server
        '''
        _LOGGER.info("Starting Commander")
        async with self._commanderLock:
            if not self.messageQueue.empty():
                await self._empty_message_queue()

            while self.connection:
                try:
                    _LOGGER.verbose(f"Waiting for message from commander queue")
                    cmd = await self.messageQueue.get()

                    if "id" not in cmd:
                        cmd["id"] = self.next_id
                    else:
                        msg_id = cmd["id"]
                        if msg_id < self.__last_id:
                            _LOGGER.debug(f"{msg_id} was already used for a websocket message. Increasing id of command {cmd}")
                            new_id = self.next_id
                            cmd["id"] = new_id
                            if msg_id in self._callback_queues:
                                _LOGGER.debug(f"Message {msg_id} has a callback. Removing object from key {msg_id} and putting it in {new_id}.")
                                self._callback_queues[new_id] = self._callback_queues.pop(msg_id)

                    send = await asyncio.wait_for(self.websocket.send(json.dumps(cmd)), timeout=10) #@IgnoreException
                    _LOGGER.debug(f"Command send")
                    _LOGGER.verbose(cmd)
                except TimeoutError:
                    if cmd["type"] == "call_service":
                        _LOGGER.warning(f"Calling service action {cmd['domain']}.{cmd['service']} timed out")
                    else:
                        _LOGGER.warning(f"Sending message {cmd['type']} timed out")
                except (TypeError, KeyError, IndexError, OSError) as exce:
                    _LOGGER.error(f"Exception occured in commander while sending command {cmd}: {exce}")
                except websockets.exceptions.ConnectionClosedError as exce:
                    _LOGGER.error(f"Commander stopped due to connection closing")
                    _LOGGER.debug(exce)
                    break
                except (asyncio.CancelledError, asyncio.exceptions.CancelledError):
                    break

        _LOGGER.error("Commander stopped")

    async def __async_ping_pong(self):
        '''
        Sends pings and receives the pongs from the Home Assistant Server
        '''
        _LOGGER.info("Starting ping-pong script")
        pong_timeout = 10
        pongs_missed = 0
        missed_max = MAX_PONGS_MISSED
        while self.connection:  
            await asyncio.sleep(self.ping_interval)  
            # self.last_id += 1
            ping_id = self.next_id
            ping_dict = {"id": ping_id, "type": "ping"}
            pong_task = self.add_callback(ping_id)

            _LOGGER.debug(f"Sending Ping id {ping_id}")
            try:
                await self.messageQueue.put(ping_dict)
                res = await asyncio.wait_for(pong_task, timeout=pong_timeout) #@IgnoreException
                ping_id = res["id"] ##Update the ping_id in case the commander increased it
                pongs_missed = 0
            except asyncio.exceptions.TimeoutError:
                pongs_missed += 1
                _LOGGER.error(f"Did not receive pong back from Home Assistant within {pong_timeout} seconds, missed {pongs_missed} pongs in a row")
                if pongs_missed >= missed_max: 
                    self.__connection = False
                    break
            except websockets.exceptions.ConnectionClosedError as exce:
                _LOGGER.error(f"Ping Pong errored due to connection closing: {exce}")
                self.__connection = False
                break

            _LOGGER.debug(f"Received pong from ping id {ping_id}")

        _LOGGER.warning("Ping pong function has stopped, Closing websocket to start reconnect")
        await self.websocket.close()
        async with self.websocketCondition:
            self.websocketCondition.notify_all()
        self.reconnect_client()

    async def _empty_message_queue(self):
        """
        Empties the queue with messages to be send to the commander, to prevent any service-actions from being performed on reconnect
        """
        while not self.messageQueue.empty():
            await self.messageQueue.get()
            self.messageQueue.task_done()

    async def _new_entities_subscribe(self, *entities):

        for ent in entities:
            if ent not in all_entities:
                ##all_entities should not have to perform a check for this.
                all_entities[ent] = {'entity_id': ent}

        if not self.stateDict or not self.commanding:
            return

        _LOGGER.debug(f"Subscribing to new entities {entities}.")

        states_msg = {
        "id": self.next_id,
        "type": "get_states"
        }

        result = await self._async_get_message_result(states_msg)
        states = result["result"]

        for ent_state in filter(lambda ent_state: ent_state["entity_id"] in entities, states):
            self._stateDict[ent_state["entity_id"]] = ent_state

        for entity in filter(lambda entity: entity not in self._stateDict, entities):
            _LOGGER.error(f"entity {entity} could not be found on the Home Assistant server. Setting it's state to unknown.")
            state = stateDictType(entity_id=entity,state="unknown", attributes={}, last_changed=None, last_reported=None,last_updated=None, context=None)
            self._stateDict[entity] = state

        ##Will remove the last_id later and have it set by the commander
        for subscr in trigger_headers(entities, last_id=self.__last_id):
            await self.messageQueue.put(subscr)
        
        last_id = subscr["id"]
        if last_id > self.__last_id:
            self.__last_id = last_id
        
        _LOGGER.info(f"Succesfully subscribed to new entities {entities}.")

    def add_callback(self, message_id : int) -> asyncio.Task:
        """
        Indicates the response of the server from the message with the specified id is needed somewhere, and puts it in an asyncio queue when received.
        The queue object is discarded from the callbacks afterwards (Since message_id's cannot be reused), and the task object returns the response.

        Parameters
        ----------
        message_id : int
            The id of the message to expect the response from

        Returns
        -------
        asyncio.Task
            A task that can be awaited on, that will return the response from the server.
        """        
        if message_id in self._callback_queues:
            _LOGGER.warning(f"A message with id {message_id} already has a callback assigned. Returning that queue")
            return self._callback_queues[message_id]
        
        def remove_queue(future : asyncio.Future):
            try:
                id = future.result()["id"]
                pop_id = id
            except (asyncio.CancelledError, asyncio.exceptions.InvalidStateError):
                pop_id = message_id
            self._callback_queues.pop(pop_id, None)

        resp_queue = asyncio.Queue(1)
        self._callback_queues[message_id] = resp_queue
        task = asyncio.create_task(resp_queue.get())
        task.add_done_callback(remove_queue)
        return task
    #endregion

    def register_new_element(self, element : elements.Element):
        
        if hasattr(element,"_HAclient"):
            element._HAclient = self

        if not hasattr(element,"entity") and not isinstance(element,HAelement):
            return
        

        if not isinstance(element,HAelement):
            HAelement.wrap_element(element,self)

        if element.HAclient == None:
            element._HAclient = self

        if not hasattr(element,"entity"):
            return
        
        entity_id = element.entity

        if entity_id not in all_entities:
            _LOGGER.debug(f"{entity_id} is not defined in config, adding it.")
            
            if self.websocket is not None:      
                self.messageQueue.put_nowait(trigger_headers(entity_id,self.__last_id))

            all_entities[entity_id] = {'entity_id': entity_id}

        if entity_id in self.elementDict:                       
            self._elementDict[entity_id].add(element)
        else:
            self._elementDict[entity_id] = set([element])

    def _add_element(self, *element_list, update_elements=True, layout_element=None, internalbatch=True):
        '''
        Adds an element that is called when it's asociated entity is updated
            Parameters:
                element_list: The elements that are to be watched by the client. 
                            Can be any pssm or HAelement. If a pssm layout, will add all elements individually. The entity_id linked to a layout is copied to their sub elements, unless said element has its own entity_id
                update_elements (bool): update_element after adding (not implemented)
                layout_element: For internal purposed mainly. The first layout element an element is associated with
                internalbatch: Mainly internal and not implemented
        '''
        for element in element_list:
            if hasattr(element, "entity"):
                
                if not isinstance(element, HAelement):
                    ##This should take care of parsing
                    HAelement.wrap_element(element,self)
                
                element._HAclient = self
                entity_id = element.entity
                
                if entity_id not in all_entities:
                    _LOGGER.info(f"{entity_id} is not defined in config, adding it.")
                    if self.websocket is not None:      
                        self.messageQueue.put_nowait(trigger_headers(entity_id,self.__last_id))

                    all_entities[entity_id] = {'entity_id': entity_id}

                if entity_id in self.elementDict:                       
                    self._elementDict[entity_id].add(element)
                else:
                    self._elementDict[entity_id] = set([element])

            if isinstance(element,elements.Layout):
                for subElement in element.createEltList():
                    self.add_element(subElement, layout_element=element, update_elements=False, internalbatch=internalbatch)
                    
    def update_element_entity(self, element : HAelement, new_entity : str, old_entity = None, update_now=True):
        """
        Updated the entity associated with the element to new_entity
        args:
            element: the element to change the entity off
            new_entity (str, list): entity_id (or list of) to associate with the element (lists may not be implemented yet-ish)
            update_now(bool): immediately update the element by calling its trigger_function, using the state currently in the state dict.
        """
        if not self.pssmScreen.printing and old_entity == None:
            self.register_new_element(element)
            return

        if old_entity:
            _LOGGER.debug(f"Updating {element} entity from {old_entity} to {new_entity}")
            if old_entity in self.elementDict:
                if element in self.elementDict[old_entity]:
                    _LOGGER.debug(f"Removing {element} from {old_entity}")
                    self._elementDict[old_entity].remove(element)

        if new_entity in self.elementDict:                       
            self._elementDict[new_entity].add(element)
        else:
            self._elementDict[new_entity] = set([element])
        if update_now:
            if new_entity not in self.stateDict:
                asyncio.create_task(self._new_entities_subscribe(new_entity))
            elif callable( func := getattr(element,"trigger_function",False)):
                to_state = self.stateDict.get(new_entity,{})
                ent_dict = {"entity_id": new_entity, "to_state": to_state, 'from_state': None, 'context': None}
                ent_dict = triggerDictType(**ent_dict)
                if asyncio.iscoroutine(func):
                    asyncio.create_task(func(element, self.stateDict[new_entity]))
                elif func:
                    asyncio.create_task(tools.wrap_to_coroutine(func,element,self.stateDict[new_entity]))
            
    def setup_entity_functions(self):
        for ent, ent_config in all_entities.items():
            if "trigger_functions" in ent_config:
                funcs = ent_config["trigger_functions"]
                if not isinstance(funcs, Sequence):
                    funcs = [funcs]
                self.add_entity_function(ent, call_functions=funcs)
        return

    def add_entity_function(self, entity_id: str, call_functions: list[tuple[Union[str,Callable]], bool]):
        '''
        Adds a function that is called when entity_id is updated. Can accept multiple trigger entities for the function

                Parameters:
                        entities (str): The entity that is associated with the function. If a list is provided, the function will be associated with all entities in it
                        call_function (function): The function to be called. Needs to accept a trigger_dict (called on entity update) and current stateDict (dict, with all current entity states, called after adding and reconnect). 
                                If called after adding, needs to accept a current state dict as well, which is none if the state is not defined yet. Set up can be call_function(power_funcs_test(trigger_dict=None,stateDict=None)
                        call_after_add (bool): after adding the function to the dict, run the function (true) or not (false). If starting up the dashboard, this will decide if the function is called after all states have been processed. 
        '''

        for func_tuple in call_functions:
            if isinstance(func_tuple, (tuple,list)):
                func, call_after_add = func_tuple
            elif isinstance(func_tuple,(dict,MappingProxyType)):
                func = func_tuple.get("function")
                call_after_add = func_tuple.get("call_on_connect", False)
            else:
                call_after_add = False
                func = func_tuple
            
            if isinstance(func,str):
                func = self.pssmScreen.parse_shorthand_function(func)

            if isinstance(entity_id,str):
                if entity_id not in all_entities:
                    _LOGGER.warning(f"{entity_id} is not defined in the configuration entities")
                if entity_id in self.functionDict:
                    self._functionDict[entity_id].append((func, call_after_add))
                else:
                    self._functionDict[entity_id] = [(func, call_after_add)]
                
                #Only calls if the function is added while connected. Otherwise it's handled in the connect script.
                if call_after_add and self.connection:
                    trigger_dict = triggerDictType(entity_id=entity_id, to_state=self.stateDict[entity_id], from_state=None, context=None)
                    func(trigger_dict, self)


    async def update_states(self,trigger):
        "Updates the clients state dict from the trigger. Calls any update functions (elements and general functions) associated with the entity as well."
        updated_entity = trigger["event"]["variables"]["trigger"]["entity_id"]
        to_state = trigger["event"]["variables"]["trigger"]["to_state"]

        trigger_dict = trigger["event"]["variables"]["trigger"]
        trigger_dict["context"] = trigger["event"]["context"]
        self._stateDict[updated_entity] = to_state
        
        trigger_dict = MappingProxyType(trigger_dict)

        coro_list = []
        func_list = []
        if updated_entity in self.functionDict:
            _LOGGER.debug(f"Updating {updated_entity} functions: {self.functionDict[updated_entity]}")
            for (func, __) in self.functionDict[updated_entity]:
                func_list.append(func)
                coro_list.append(tools.wrap_to_coroutine(func, trigger_dict, self))

        if self.connection and self.authenthicated:
            if not self.updatingAll:
                ent_elts = self.elementDict.get(updated_entity,[])
                _LOGGER.debug(f"Updating {updated_entity} elements: {ent_elts}")
                for element in ent_elts:
                    element : HAelement
                    if not hasattr(element,"trigger_function"):
                        continue
                    func = element.trigger_function
                    coro_list.append(tools.wrap_to_coroutine(func,element,trigger_dict))
                    func_list.append(func)
            else:
                _LOGGER.warning("Wanted to update {} but update all in progress".format(updated_entity))
        
        if coro_list:
            L = await asyncio.gather(*coro_list, return_exceptions=True) #, return_exceptions=True)
            for i, res in enumerate(L):
                if isinstance(res,Exception): 
                    _LOGGER.error(f"{func_list[i]} returned an exception: {res} ")

    async def subscribe_to_trigger(self, entity: Union[str,list[str]] = None, trigger: dict = None):
        if entity == None and trigger == None:
            return

        if entity:
            header = trigger_headers(entity, self.__last_id)
        else:
            header = trigger
        
        if self.commanding:
            t = self.add_callback(header["id"])
            await self.messageQueue.put(header)
            subscr_resp = await t
        else:
            await self.websocket.send(json.dumps(header))
            subscr_resp = {}

            while subscr_resp.get("id", None) != header["id"]:
                subscr_resp = await self.websocket.recv()
                subscr_resp = json.loads(subscr_resp)

                if subscr_resp.get("type", None) == "event":
                    try:
                        asyncio.create_task(self.update_states(subscr_resp))
                    except (TypeError, KeyError, IndexError, OSError) as exce:
                        _LOGGER.error(f"Error in update states for {subscr_resp}: {exce}")

        _LOGGER.debug(subscr_resp)
        if not "success" in subscr_resp:
            _LOGGER.error(f"Error with a subscribe header {subscr_resp}")
        elif not subscr_resp["success"]:
            _LOGGER.error(f'Failed to subscribe to {header["trigger"]["entity_id"]}, server responded with {subscr_resp}')
        else:
            for func in self._subcribe_callbacks:
                func(self, header)
            return header

        return subscr_resp

    def call_service(self, elt : Union[HAelement,elements.Element] = None, coords = None, service = None, service_data = None, target = None, response = False, *args, **kwargs):
        """
        Placeholder for the old terminology of service (which was renamed to service_action in HA)

        Parameters
        ----------
        elt : Union[HAelement,elements.Element], optional
            element to optionally provide data. Will be deprecated as call_service should obtain this data from the element, by default None
        coords : _type_, optional
            the coordinates where the click was registered (not used but required to make it work as a tap_action), by default None
        service : _type_, optional
            service to call. Overwrites data from element if one has been passed, by default None
        service_data : _type_, optional
            optional data to send along with the service. Overwrites service data from element if one has been passed, by default None
        target : _type_, optional
            target to call the service on. Overwrites element target is one has been passed., by default None
        response_function : _type_, optional
            function to return the response from the service call too. This will also tell Home Assistant to return a response from this service call. Must accept a response argument. If called from an element click, this function should also accept an elt and coords parameter., by default None
        """
        return self.call_service_action(elt,coords,service,service_data,target, response, **kwargs)

    def call_service_action(self, elt : Union[HAelement,elements.Element]=None, coords : screen.CoordType = None, 
                            action : str = None, action_data : Union[dict,str] = None, target : Union[dict,str,None] = None, response : bool = False, service_id : Optional[str] = None,
                            *args, **action_data_kwargs) -> Union[asyncio.Task[Any,dict],None]:
        '''
        Builds a message for the commander to call a service action on the Home Assistant server. Mainly made to work as the tap_action function of PSSM elements, but can also be called separately
        Calls a home assistant service action
        Parameters:
                elt: element to optionally provide data. Will be deprecated as call_service should obtain this data from the element
                coords: the coordinates where the click was registered (not used but required to make tap_action work)
                action: service action to call. Overwrites data from element if one has been passed
                action_data: optional data to send along with the action. Overwrites action data from element if one has been passed
                target: target (entity id, area etc.) to call the action on. Overwrites element target is one has been passed.
                response_function (bool): If true, the response of the action call will be caught and forwarded via the task
                response (bool): If true, the function tell the client to catch the result of the action call. In this case, the `call_service_action` function will return an asyncio Task that can be awaited, which will return the result of the response (or the entire response if the call was unsuccesfull)
        '''
        _LOGGER.debug("Calling a service action")
        service_header = {}
        if service_id != None and action == None:
            _LOGGER.warning("Got a service id but this is not yet implemented from the config script")
            action = service_id

        if elt != None and hasattr(elt,"_serviceCallTime"):
            elt.update(updateAttributes={"_serviceCallTime": datetime.now()}, skipPrint=True, skipGen=True)

        return_response = response
            
        if action is not None:
            if action_data == None and action_data_kwargs:
                action_data = action_data_kwargs
            if target == None: target = getattr(elt,"entity",None)
            
            service_header = self.build_service_header(action_id = action, action_data = action_data, target = target, return_response=return_response)
        
        if return_response and service_header:
            response_task = self.loop.create_task(self._async_get_message_result( service_header, elt, coords))
            _LOGGER.debug("Made task for service callback and added to queue")
            return response_task
        else:
            asyncio.run_coroutine_threadsafe(
                self.messageQueue.put(service_header),
                self.pssmScreen.mainLoop
            )
        return

    def build_service_header(self, action_id : str, action_data : Optional[dict], target : Optional[Union[dict,str]] = None, return_response : bool=False) -> actionCallDict:
        '''
        Builds the header to send to Home Assistant to call the asked for service with the service data and target ID
            Parameters:
                action_id: the service to call, as you would call it in home assistant (i.e. light.turn_on)
                action_data: optional data to send along with the service
                target: target to call the service on
                return_response: tell Home Assistant to return a response from this service call
        '''
        if action_id in all_service_actions:
            service_config = all_service_actions[action_id]
            full_service = service_config["action"]
        else:
            full_service = action_id
            service_config = {}
        
        full_service = full_service.split(".")
        try:
            serv_domain = full_service[0]
            serv_service = full_service[1]
        except (TypeError, IndexError):
            _LOGGER.warning(f"Error in service action call {full_service}", exc_info=True)
            return

        msg_id = self.next_id
        service_header = { 
            "id": msg_id,
            "type": "call_service",
            "domain": serv_domain,
            "service": serv_service,
            }
        if action_data == None:
            if "data" in service_config:
                service_header.update({"service_data" : service_config["service_data"]})
        else:
            service_header.update({"service_data" : action_data})

        if target == None:
            if "target" in service_config:
                service_header.update({"target": service_config["target"]})
        else:
            if isinstance(target,(str,list,tuple)):
                target = {"entity_id": target}
            service_header.update({"target": target})
        
        if "entity_id" in service_header.get("target",{}):
            ent_target = service_header["target"]["entity_id"]
            if isinstance(ent_target,str) and ENTITY_TAG_KEY in ent_target:
                service_header["target"]["entity_id"] = parse_entity_tag(ent_target)
            elif isinstance(ent_target,(tuple,list)):
                targets = list(ent_target)
                for i, ent in enumerate(targets):
                    if ENTITY_TAG_KEY in ent:
                        targets[i] = parse_entity_tag(ent)
                service_header["target"]["entity_id"] = targets

        service_header.update({"return_response": bool(return_response)})
        return service_header

    def build_element_service_header(self, elt : Union[elements.Element, HAelement]) -> actionCallDict:
        """
        Builds a service header from a PSSM element; 

        Parameters
        ----------
        elt : Union[elements.Element, HAelement]
            the element to build from

        Returns
        -------
        actionCallDict
            The dict to put in the message queue to send to Home Assistant
        """

        if not hasattr(elt,"service_action"):
            msg = f"Calling a service from an element requires a service_action attribute to be set. {elt.id} does not have one."
            _LOGGER.exception(AttributeError(msg), stack_info=True)
            return

        service_id = elt.service
        
        service_data = {}
        if hasattr(elt,"service_data"):
            service_data = elt.service_data
        
        if hasattr(elt,"service_data_map"):
            elt.service_data_map
            for k,v in elt.service_data_map.items():
                if hasattr(elt,v): service_data[k] = getattr(elt,v)

        if not service_data: service_data = None

        target = None
        if service_data != None and "target" in service_data:
            target = service_data["target"]
            service_data.pop("target")
        elif hasattr(elt,"entity"):
            target = {"entity_id": elt.entity}

        service_message = self.build_service_header(service_id, service_data, target)
        return service_message

    async def _async_get_message_result(self, message, *args) -> Union[Any,Literal[False]]:
        """
        Quick hand to easily get the response of a message send to the command.
        Simply 

        Parameters
        ----------
        message : dict
            The message to put in the message queue. If it does not have an id yet, the id will be set automatically

        Returns
        -------
        dict
            The response. If the reponse was successful, it just returns the ["result"]["response"] part of the response (Like you would see in the developer_tools perform action part of the frontend). Returns False if the response was not succesful.
        """        

        if "id" not in message:
            message_id = self.next_id
            message["id"] = message_id
        else:
            message_id = message["id"]

        task = self.add_callback(message_id)

        #self.next_message = service_header
        await self.messageQueue.put(message)
        _LOGGER.debug(f"Waiting for service response from id {message_id}")
        response = await task
        _LOGGER.debug(f"Message {message_id} returned service response.")
        if not response.get("success",False):
            _LOGGER.error(f"Error getting a service action response. Received {response}")
        return response

    async def client_update_elements(self,entity_id=None,internalbatch=False,update_all=False, timeout : int = None) -> Optional[list[Coroutine]]:
        """
        Updates the elements associated with entity_id to the last received state. Does not handle triggers, so trigger_function is called with 'old_state': False

        Parameters
        ----------
        entity_id : _type_, optional
            The entity_id to update all connected elements of, by default None
        internalbatch : bool, optional
            Start a _batch_writing process, by default False
        update_all : bool, optional
            Update all elements connected to entity, by default False
        timeout : int, optional
            If updating all elements, set an optional timeout, by default None
            The update tasks will continue in the background, but it prevents the function from blocking, i.e. when connecting

        Returns
        -------
        Optional[list[Coroutine]]
            If update_all is False, a list of all element's trigger functions is returned.
            Can be used in i.e. an asyncio.gather call.
        """        
        
        "Updates the elements associated with entity_id to the last received state. Does not handle triggers, so trigger_function is called with 'old_state': False"
        #InternalBatch: defaults to true. If true, will start a batch writing in the function. Otherwise, will asume it is defined outside of it.
        #If state_dict is still empty/does not exist: make it sleep for a bit

        ##This has to be rewritten a bit to accept async functions
        if internalbatch: self.pssmScreen.start_batch_writing()
        if update_all:
            _LOGGER.debug("[HAClient]: Updating all elements")
            coro_list = []
            self.updatingAll = True
            self.pssmScreen.start_batch_writing()
            for entity in all_entities:
                coro_list.extend(
                    await self.client_update_elements(entity_id=entity,internalbatch=False))
            if coro_list:
                try:
                    done, pending = await asyncio.wait(coro_list,loop=self.pssmScreen.mainLoop,timeout=timeout)
                except Exception as exce:
                    _LOGGER.error(exce)
                    pass
                
                if pending:
                    msgs = set()
                    for task in pending:
                        task : asyncio.Task
                        coro = task.get_coro()
                        vars = coro.cr_frame.f_locals
                        if "self" in vars:
                            elt = vars["self"]
                            func = getattr(elt,"trigger_function",None)
                        elif "element" in vars:
                            elt = vars["element"]
                            func = getattr(elt,"trigger_function",None)
                        else:
                            elt = getattr(vars,"args", [None])[0]
                            func = getattr(vars,"func",None)
                        
                        if callable(func):
                            if "HomeAssistantClient" in func.__module__:
                                func = func.__name__
                            else:
                                func = f"{func.__module__}.{func.__name__}"
                        
                        if elt == None:
                            msg = func
                        else:
                            msg = f"{elt}: {func}"
                        
                        if msg != None:
                            msgs.add(msg)

                    _LOGGER.warning(f"Element updates are taking a long time: {msgs}. Continuing in background.")
                
                    for task in done:
                        task : asyncio.Task
                        if task.exception() != None:
                            coro = task.get_coro()
                            vars = getattr(coro.cr_frame,"f_locals", {})
                            if "self" in vars:
                                elt = vars["self"]
                            elif "element" in vars:
                                elt = vars["element"]
                            else:
                                elt = getattr(vars,"args", [None])[0]

                            _LOGGER.warning(f"Element {elt} raised an error in it's trigger_function {getattr(coro,'__qualname__','unknown_function_name')}")
                    

            self.updatingAll = False
            self.pssmScreen.stop_batch_writing()
            return
        else:
            try:
                coro_list = []
                if entity_id in self.elementDict:
                    for element in self.elementDict[entity_id]:
                        element : HAelement
                        func = False
                        to_state = self.stateDict[entity_id]
                        ent_dict = {"entity_id": entity_id, "to_state": to_state, 'from_state': None, 'context': None}
                        ent_dict = triggerDictType(**ent_dict)

                        if isinstance(element,elements.Slider) and hasattr(element,"trigger_function"):
                            #The slider update would jump around a bit since the lights fade. It updates on touch, and then has a delayed callback to update the indicator precisely
                            #The delay is thus not called if the service was not called recently (so when fading it from your phone eg)
                            
                            ##This should be moved to the slider  trigger_function itself
                            if element.serviceCallTime != None and (datetime.now() - element.serviceCallTime).total_seconds() < 5:
                                self.loop.create_task(self.__async_update_later(entity_id=entity_id, element=element))
                            else:
                                func = element.trigger_function
                        elif hasattr(element,"trigger_function"):
                            func = element.trigger_function
                            if isinstance(element,elements.Icon):
                                if element.fileError:
                                    msg = " Error updating {} icon for state {}: image {} does not exist".format(entity_id,to_state,element.icon)
                                    _LOGGER.warning(msg)

                        else:
                            _LOGGER.warning("{}: Wanted to update unknown element type: {}".format(entity_id, element))
                    
                        if func:
                            if asyncio.iscoroutinefunction(func):
                                coro_list.append(func(element,ent_dict))
                            else:
                                coro_list.append(asyncio.to_thread(
                                        func, element, ent_dict))
                    if internalbatch and not self.updatingAll: self.pssmScreen.stop_batch_writing()
                return coro_list
            except FuncExceptions as exce:
                msg = exce
                _LOGGER.error(f"Caught error updating elements: {exce}")
                _LOGGER.debug(msg)
                return []
        
    async def __async_update_later(self, entity_id, element, wait_time=5):
        await asyncio.sleep(wait_time)
        entity_range = element.range
        if hasattr(element,"attribute"):
            try:
                cur_state = self.stateDict[entity_id]["attributes"]["brightness"]
            except KeyError:
                cur_state =  self.stateDict[entity_id]
        else:
            cur_state =  self.stateDict[entity_id]
        if cur_state == "off":
            cur_state_perc = 0
        else:
            try:
                cur_state_perc = int(((cur_state - entity_range[0])/(entity_range[1] - entity_range[0]))*100)
            except (TypeError,IndexError):
                cur_state_perc = 0
        if abs(element.position - cur_state_perc) > 5:
            element.trigger_function(element,self.stateDict[entity_id])

    def get_unit(self, unit_type : Literal["length","accumulated_precipitation","mass","pressure","temperature", "volume", "wind_speed"]) -> Optional[str]:
        """
        Returns the unit that should be the main one used for this Home Assistant instance connected to the provided unit_type.
        If the unit_type cannot be found, returns None
        Generally, if an entity has its own unit_of_measurement attribute e.g. (a weather entity has multiple unit attributes for example), it is best to use said attribute.

        Parameters
        ----------
        unit_type : Literal[&quot;length&quot;,&quot;accumulated_precipitation&quot;,&quot;mass&quot;,&quot;pressure&quot;,&quot;temperature&quot;, &quot;volume&quot;, &quot;wind_speed&quot;]
            Which type of unit to get the unit of. I.e. for a distance travelled state, use `unit_type` = length
            (So, the physical quantity type)

        Returns
        -------
        str
            The unit of the requested unit_type. None if it does not exist.
        """

        units = self._HAconfig["unit_system"]
        return units.get(unit_type,None)


class dummyClient:
    '''
    Dummy client for testing. Does not interact with pssm and likely misses functions.
    Does have some additional functions to easily get and read out responses.
    Main usage is by using the function dict to get entity triggers
    '''
    def __init__(self):
        _LOGGER.info("Starting dummy client")
        self._websocket = None
        self.hass_data = CORE.config.configuration["home_assistant"]
        self.__last_id = 0
        self.stateDict = {}

        self.__connection = False
        self.listenerTask = DummyTask()
        self.commanderTask = DummyTask()
        self.pingpongTask = DummyTask()
        self.message_queue : asyncio.Queue = False
        "queue with messages to send to the server"

        self._callback_queues : dict[int, asyncio.Queue] = {}
        """
        Dict with [message_id]: asyncio.Queue; When a message is returned with the corresponding id, the message is put in the queue. Used to get e.g. service responses
        In general, use this only to signify a callback. The Queue
        """

        self.entity_queues = {}
        "Dict with [entity_id]: asyncio.Queue; when the entity is updated, the trigger dict is put in the queue"

        self.functionDict = {}
        "Dict with [entity_id]: function; calls the function when the entity updates, and on connecting"

    @property
    def next_id(self) -> int:
        "Returns the next_id that can be used for a message (presumably)"
        self.__last_id += 1
        return self.__last_id

    async def connect_client(self, url, auth_token):
        """Starts the function that connects to Home Assistant""" 
        self.loop = asyncio.get_event_loop()
        await self.__async__connect(url, auth_token)

    async def start_client(self):
        """Starts the infinite loops to manage the Home Assistant connection (listener, commander and ping-pong)"""
        _LOGGER.debug("Starting Dummy HA Client")
        if self.listenerTask.done():
                self.listenerTask = self.loop.create_task(self.__async_listen())
        if self.commanderTask.done():
                self.commanderTask = self.loop.create_task(self.__async_command())

    async def __async__connect(self, url: str, auth_token: str):
        """Sets up a websocket connection to Home Assistant"""
        
        hass_data = self.hass_data
        uri = "ws://{}/api/websocket".format(url)
        token = auth_token
        auth_header =    {
            "type": "auth",
            "access_token": token     }
        _LOGGER.info("Attempting connection to {}".format(uri))

        # perform async connect, and store the connected WebSocketClientProtocol
        # object, for later reuse for send & recv
        try:
            self._websocket = await websockets.connect(uri)           
            await self.websocket.recv()
            self.__connection = True
            await self.websocket.send(json.dumps(auth_header))
            auth_res = json.loads(await self.websocket.recv())
            if auth_res["type"] == "auth_ok":
                self.authenthicated = True
                _LOGGER.info(f"Connected to Home Assistant {auth_res}")
            else:
                _LOGGER.error(f"Authentication failed {auth_res}")
                pass

            #This gets literally ALL states, but it seems to handle fine. May however cause problems in large setups.
            states_header = {"id": self.next_id, "type": "get_states" } 

            await self.websocket.send(json.dumps(states_header))
            all_states = json.loads(await self.websocket.recv())
            if all_states["success"]:
                #Finding the corresponding entitiy's entry in the list of all states
                _LOGGER.debug("Received all states from Home Assistant")
                initial_states = list(filter(lambda entity: entity["entity_id"] in all_entities, all_states["result"])) 
                all_states = None
                initial_dict = {}
                for entity in initial_states:
                    initial_dict[entity["entity_id"]] = entity

                self.stateDict = initial_dict
                
                _LOGGER.debug(f"Updating functions in function dict")
                called_functions = []
                for func_entity in self.functionDict:
                    for func_tuple in self.functionDict[func_entity]:
                        (func, call_func) = func_tuple
                        if func not in called_functions and call_func:
                            try:
                                func(state_dict=self.stateDict)
                                called_functions.append(func)
                            except FuncExceptions as exce:
                                _LOGGER.warning(f"Error calling function {func} for entity {func_entity}: {exce}, removed from function dict")
                                self.functionDict[func_entity].remove(func_tuple)
            else:
                _LOGGER.error(f"Failed to get states {all_states}")

            subscribe_headers = trigger_headers(all_entities,self.__last_id)
            subscribe_fails = 0
            for header in subscribe_headers:
                await self.websocket.send(json.dumps(header))
                subscr_resp = await self.websocket.recv()
                subscr_resp = json.loads(subscr_resp)
                _LOGGER.debug(subscr_resp)
                if not "success" in subscr_resp and subscr_resp["type"] != "event":
                    _LOGGER.error(f"Error with a subscribe header {subscr_resp}")
                elif subscr_resp["type"] != "event":
                    if not subscr_resp["success"]:
                        _LOGGER.error(f'Failed to subscribe to {header["trigger"]["entity_id"]}, server responded with {subscr_resp}')
                        subscribe_fails += 1
            if subscribe_fails == 0:
                _LOGGER.info("Succesfully subscribed to all entities")

            self.__last_id = header["id"]
        except TimeoutError:
            _LOGGER.error(f'Failed to establish connection to {self.hass_data["url"]}')
        except FuncExceptions as exce:
            _LOGGER.error(f"Couldn't connect {exce}")
        return        
    
    async def __async_listen(self):
        '''
        Listens to messages from the Home Assistant server
        '''
        _LOGGER.info("Starting Listener")
        #while self.connection:
        self.listening = True
        try:
            async for message in self.websocket:
                message = json.loads(message)
                id = message["id"]
                _LOGGER.verbose(f"Received message {id}")
                _LOGGER.verbose(message)
                if id in self._callback_queues:
                    ##Maybe have a seperate queue for service responses and one for pings/pongs.
                    queue = self._callback_queues[id]
                    _LOGGER.debug(f"Putting response from message id {id} in queue")
                    try:
                    # attempt to add an item
                        queue.put_nowait(message)
                    except asyncio.QueueFull:
                        _LOGGER.debug(f"Message id {id} has already queued a response")
                
                if message.get("type") == "event":
                    # [ ]: remove this try once the illegal instruction error is fixed
                    _LOGGER.debug(message)
                    ent = message["event"]["variables"]["trigger"].get("entity_id", False)
                    if ent and ent in self.entity_queues:
                        await self.entity_queues[ent].put(message)
                elif message.get("type") == "result":
                    if not message.get("succes", True):
                        _LOGGER.warning("Unsuccesful request: {}".format(message))
                        self.update_Icon()
                ##Catch all other message types that are not pongs
                elif message.get("type") != "pong":
                    _LOGGER.warning("Message seems unfamiliar: {}".format(message))
        
        except websockets.exceptions.ConnectionClosedError as exce:
            _LOGGER.error(f"Websocket connection closed, calling reconnect {exce}")
            self.__connection = False
            self.reconnect_client()
                        
        _LOGGER.warning("Listener stopped")

    async def __async_command(self):
        #Pick the first entry every iteration and remove it from the list after the send --> maybe but hasn't lead to problems yet though
        '''
        Sends messages put into HAclient.message_queue to the Home Assistant server
        '''
        _LOGGER.info("Starting Commander")
        self.message_queue = asyncio.Queue()
        while self.connection:
            try:
                _LOGGER.verbose(f"Waiting for message from commander queue")
                cmd = await self.message_queue.get()

                if "id" not in cmd:
                    cmd["id"] = self.next_id
                else:
                    msg_id = cmd["id"]
                    if msg_id < self.__last_id:
                        _LOGGER.debug(f"{msg_id} was already used for a websocket message. Increasing id of command {cmd}")
                        new_id = self.next_id
                        cmd["id"] = new_id
                        if msg_id in self._callback_queues:
                            _LOGGER.debug(f"Message {msg_id} has a callback. Removing object from key {msg_id} and putting it in {new_id}.")
                            self._callback_queues[new_id] = self._callback_queues.pop(msg_id)

                send = await asyncio.wait_for(self.websocket.send(json.dumps(cmd)), timeout=10)
                _LOGGER.debug(f"Command send")
                _LOGGER.verbose(cmd)
            except TimeoutError:
                _LOGGER.warning("Calling {}.{} timed out".format(cmd["domain"],cmd["service"]))
            except FuncExceptions as exce:
                _LOGGER.error(f"Exception occured in commander:{exce}")

        _LOGGER.error("Commander stopped")

    def add_callback(self, message_id : int) -> asyncio.Task:
        """
        Indicates the response of the server from the message with the specified id is needed somewhere, and puts it in an asyncio queue when received.
        The queue object is discarded from the callbacks afterwards (Since message_id's cannot be reused), and the task object returns the response.

        Parameters
        ----------
        message_id : int
            The id of the message to expect the response from

        Returns
        -------
        asyncio.Task
            A task that can be awaited on, that will return the response from the server.
        """        
        if message_id in self._callback_queues:
            _LOGGER.warning(f"A message with id {message_id} already has a callback assigned. Returning that queue")
            return self._callback_queues[message_id]
        
        resp_queue = asyncio.Queue(1)
        self._callback_queues[message_id] = resp_queue
        task = asyncio.create_task(resp_queue.get())
        return task

    async def get_response(self,message_dict : dict) -> dict:
        """
        Send a command to the server and return the response.
        see https://developers.home-assistant.io/docs/api/websocket for possible commands
        args:
            message_dict: dict with the message. Optionally add id, or let the client automatically set a correct one.
        """
        if "id" in message_dict:
            _LOGGER.debug("Using preset id")
            id = message_dict["id"]
        else:
            # self.last_id += 1
            id = self.next_id
            message_dict["id"] = id
        
        queue = asyncio.Queue(1)
        self._callback_queues[id] = queue
        await self.message_queue.put(message_dict)
        resp = await queue.get()
        return resp
    
    async def get_trigger_dict(self, entity : str) -> dict:
        """
        Returns the trigger dict upon the first state change of entity
        args:
            entity (str): entity_id to get the trigger from
        """
        if entity not in self.stateDict:
            _LOGGER.error(f"Client is not subscribed to {entity}")
            raise ValueError("Invalid entity")
        
        queue = asyncio.Queue(1)
        self.entity_queues[entity] = queue
        trigger = await queue.get()
        self.entity_queues.pop(entity)
        return trigger